import os
import re
import argparse

import matplotlib.pyplot as plt
import pandas as pd
import numpy as np


SMALL_SIZE = 10
MEDIUM_SIZE = 15
BIGGER_SIZE = 18

plt.rc("font", size=SMALL_SIZE)  # controls default text sizes
plt.rc("axes", titlesize=MEDIUM_SIZE)  # fontsize of the axes title
plt.rc("axes", labelsize=MEDIUM_SIZE)  # fontsize of the x and y labels
plt.rc("xtick", labelsize=SMALL_SIZE)  # fontsize of the tick labels
plt.rc("ytick", labelsize=SMALL_SIZE)  # fontsize of the tick labels
plt.rc("legend", fontsize=MEDIUM_SIZE)  # legend fontsize
plt.rc("figure", titlesize=BIGGER_SIZE)  # fontsize of the figure title


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--root", type=str, required=True)
    parser.add_argument("--output", type=str, default="results")
    args = parser.parse_args()

    REGEX_PATTERN = "(\d+)-model-(\d+)-node-exp"
    res = re.match(REGEX_PATTERN, args.root)
    n_model, n_nodes = res.groups()

    # Plot data
    # X: Time
    # Y: Succ. Rate

    sampling_time = []
    sampling_succ_rate = []

    grad_time = []
    grad_succ_rate = []

    proxy_time = []
    proxy_succ_rate = []

    for f in os.listdir(args.root):
        if f.endswith(".csv") and f != "model_info.csv":
            data = pd.read_csv(os.path.join(args.root, f))

            sampling_time.append(data["sampling-time"].mean())
            sampling_succ_rate.append(data["sampling-succ"].mean())
            grad_time.append(data["grad-time"].mean())
            grad_succ_rate.append(data["grad-succ"].mean())
            proxy_time.append(data["proxy-time"].mean())
            proxy_succ_rate.append(data["proxy-succ"].mean())
        elif f == "model_info.csv":
            data = pd.read_csv(os.path.join(args.root, f))
            gentime = data["gen_time"]
            print(f"{gentime.mean()=}, {gentime.min()=}, {gentime.max()=}")

    def sort_by_time(time, succ_rate):
        time = np.array(time)
        succ_rate = np.array(succ_rate)
        sort_idx = time.argsort()
        return time[sort_idx], succ_rate[sort_idx]

    # sort succ rate by time
    sampling_time, sampling_succ_rate = sort_by_time(sampling_time, sampling_succ_rate)
    grad_time, grad_succ_rate = sort_by_time(grad_time, grad_succ_rate)
    proxy_time, proxy_succ_rate = sort_by_time(proxy_time, proxy_succ_rate)

    fig, (ax1, ax2) = plt.subplots(
        2, 1, figsize=(8, 4), sharex=True, gridspec_kw={"height_ratios": [2.5, 1]}
    )
    fig.subplots_adjust(hspace=0.1)

    for ax in [ax1, ax2]:
        ax.plot(
            sampling_time * 1000,
            sampling_succ_rate,
            markersize=10,
            marker="x",
            linestyle="--",
            label="Sampling",
        )
        ax.plot(
            grad_time * 1000,
            grad_succ_rate,
            marker="*",
            markersize=10,
            linestyle="--",
            label="Gradient",
        )
        ax.plot(
            proxy_time * 1000,
            proxy_succ_rate,
            marker="^",
            markersize=10,
            linestyle="--",
            label="Gradient (Proxy Deriv.)",
        )
        ax.grid(True, linestyle="--", linewidth=0.5)

    ax1.set_ylim(0.66, 1)
    ax2.set_ylim(0.29, 0.43)

    ax1.spines.bottom.set_visible(False)
    ax2.spines.top.set_visible(False)
    ax1.xaxis.tick_top()
    ax1.tick_params(labeltop=False)  # don't put tick labels at the top
    ax2.xaxis.tick_bottom()

    d = 0.5  # proportion of vertical to horizontal extent of the slanted line
    kwargs = dict(
        marker=[(-1, -d), (1, d)],
        markersize=12,
        linestyle="none",
        color="k",
        mec="k",
        mew=1,
        clip_on=False,
    )
    ax1.plot([0, 1], [0, 0], transform=ax1.transAxes, **kwargs)
    ax2.plot([0, 1], [1, 1], transform=ax2.transAxes, **kwargs)

    plt.legend(loc="lower right")
    ax1.set_yticks([0.7, 0.8, 0.9, 1])
    ax2.set_yticks([0.3, 0.4])

    ax2.set_xlabel("Avg. Searching Time (millisecond)", fontweight="bold")
    ax2.xaxis.set_label_coords(0.5, 0.05, transform=fig.transFigure)
    ax2.set_ylabel("Success Rate", fontweight="bold")
    ax2.yaxis.set_label_coords(0.05, 0.5, transform=fig.transFigure)

    plt.savefig(os.path.join(args.output, f"input-search{n_model}-{n_nodes}.pdf"))
    plt.savefig(os.path.join(args.output, f"input-search{n_model}-{n_nodes}.png"))



import numpy as np
import pytest

from nnsmith.graph_gen import model_gen
from nnsmith.materialize import TestCase
from nnsmith.materialize.tensorflow import TFModelCPU, tf_dict_from_np
from nnsmith.narrow_spec import auto_opset

TestCase.__test__ = False  # supress PyTest warning


def test_onnx_load_dump(tmp_path):
    d = tmp_path / "test_onnx_load_dump"
    d.mkdir()

    gen = model_gen(
        opset=auto_opset(TFModelCPU),
        seed=54341,
        max_nodes=5,
    )

    ir = gen.make_concrete()

    model = TFModelCPU.from_gir(ir)

    model.refine_weights()  # either random generated or gradient-based.
    oracle = model.make_oracle()

    testcase = TestCase(model, oracle)
    testcase.dump(root_folder=d)

    loaded_testcase = TestCase.load(model_type=type(model), root_folder=d)

    def compare_two_oracle(src, loaded):
        assert len(loaded.input) == len(src.input)
        assert len(loaded.output) == len(src.output)
        for k, v in loaded.input.items():
            assert np.allclose(v, src.input[k], equal_nan=True)
        for k, v in loaded.output.items():
            assert np.allclose(v, src.output[k], equal_nan=True)

    # check oracle
    compare_two_oracle(oracle, loaded_testcase.oracle)

    loaded_model = loaded_testcase.model
    rerun_oracle = loaded_model.make_oracle(tf_dict_from_np(oracle.input))
    compare_two_oracle(oracle, rerun_oracle)



from __future__ import annotations

import logging
import os
import pickle
from abc import ABC
from os import PathLike
from typing import Callable, Dict, List, Type

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"
logging.getLogger("tensorflow").setLevel(logging.ERROR)
logging.getLogger("absl").disabled = True

import numpy as np
import tensorflow as tf  # type: ignore
from multipledispatch import dispatch  # type: ignore


def configure_tf_gpu_mem(max_megabytes=None):
    for gpu in tf.config.list_physical_devices("GPU"):
        tf.config.experimental.set_memory_growth(gpu, True)
        if isinstance(max_megabytes, int):
            tf.config.set_logical_device_configuration(
                gpu, [tf.config.LogicalDeviceConfiguration(memory_limit=max_megabytes)]
            )


configure_tf_gpu_mem()

from nnsmith.abstract.op import AbsOpBase, AbsTensor
from nnsmith.gir import GraphIR
from nnsmith.materialize import Model, Oracle
from nnsmith.materialize.tensorflow.forward import ALL_TF_OPS
from nnsmith.materialize.tensorflow.tfnet import TFNet
from nnsmith.util import register_seed_setter

TFNetCallable = Callable[..., Dict[str, tf.Tensor]]


@dispatch(dict)
def randn_from_specs(specs: Dict[str, tf.TensorSpec]) -> Dict[str, tf.Tensor]:
    return {
        name: tf.cast(tf.random.normal(shape=spec.shape), dtype=spec.dtype)
        for name, spec in specs.items()
    }


def np_dict_from_tf(x: Dict[str, tf.Tensor]) -> Dict[str, np.ndarray]:
    return {key: np.array(value.numpy()) for key, value in x.items()}


def tf_dict_from_np(x: Dict[str, np.ndarray]) -> Dict[str, tf.Tensor]:
    return {key: tf.convert_to_tensor(value) for key, value in x.items()}


class EagerModeCtx:
    def __init__(self, eagerly: bool) -> None:
        assert isinstance(
            eagerly, bool
        ), f"argument eagerly should not be {eagerly.__class__}. It must be a boolean."
        self.eagerly = eagerly

    def __enter__(self) -> None:
        self.old_mode = tf.config.functions_run_eagerly()
        tf.config.run_functions_eagerly(self.eagerly)

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        tf.config.run_functions_eagerly(self.old_mode)


class TFModel(Model, ABC):  # Don't directly instantiate this class
    """Wrapper class of TFNet (tf.Module)
    It only stores net. Other information like I/O info are temporarily generated.
    Other values like I/O values should be managed by the user.
    """

    def __init__(self, ir: GraphIR) -> None:
        super().__init__()
        self.ir = ir
        self.net = TFNet(ir=ir)

    @classmethod
    def from_gir(cls: Type["TFModel"], ir: GraphIR, **kwargs) -> "TFModel":
        return cls(ir)

    @property
    def version(self) -> str:
        return tf.__version__

    @property
    def native_model(self) -> TFNet:
        return self.net

    @property
    def input_like(self) -> Dict[str, AbsTensor]:
        return {ik: self.ir.vars[ik] for ik in self.ir.input_var()}

    @property
    def output_like(self) -> Dict[str, AbsTensor]:
        return {ok: self.ir.vars[ok] for ok in self.ir.leaf_var()}

    @property
    def input_specs(self) -> Dict[str, tf.TensorSpec]:
        ret: Dict[str, tf.TensorSpec] = {}
        for k, aten in self.input_like.items():
            ret[k] = tf.TensorSpec(aten.shape, aten.dtype.tensorflow(), k)
        return ret

    @staticmethod
    def name_suffix() -> str:
        return ""

    def make_oracle(
        self, inputs: Dict[str, tf.Tensor | tf.TensorSpec] = None
    ) -> Oracle:
        if inputs is None or isinstance(list(inputs.values())[0], tf.TensorSpec):
            input_dict = self.random_inputs()
        else:
            input_dict = inputs
        output_dict = self.run_eagerly(input_dict)

        input_dict = np_dict_from_tf(input_dict)
        output_dict = np_dict_from_tf(output_dict)

        return Oracle(input_dict, output_dict, provider="tf[cpu] eager")

    def dump(self, path: PathLike = "saved_tfmodel") -> None:
        os.makedirs(path, exist_ok=True)
        # gir.pkl
        with open(os.path.join(path, TFModel.gir_name()), "wb") as f:
            pickle.dump(self.ir, f)
        # tfnet
        with self.device:
            concrete_net = self.concrete_net(self.input_specs)
            tf.saved_model.save(
                self.net,
                os.path.join(path, TFModel.tfnet_dir_name()),
                signatures=concrete_net,
            )

    def dump_with_oracle(
        self,
        path: PathLike,
        inputs: Dict[str, tf.Tensor | tf.TensorSpec] = None,
    ) -> None:
        self.dump(path)
        oracle = self.make_oracle(inputs)
        oracle.dump(os.path.join(path, Oracle.name()))

    @classmethod
    def load(cls, path: PathLike) -> "TFModel":
        with open(os.path.join(path, cls.gir_name()), "rb") as f:
            gir: GraphIR = pickle.load(f)
        model = cls(gir)
        model.net = tf.saved_model.load(os.path.join(path, TFModel.tfnet_dir_name()))
        return model

    @staticmethod
    def gir_name() -> str:
        return "gir.pkl"

    @staticmethod
    def in_out_pkl_name():
        return "in_out.pkl"

    @staticmethod
    def tfnet_dir_name():
        return "tfnet"

    def random_inputs(self) -> Dict[str, tf.Tensor]:
        return {
            spec.name: tf.cast(
                tf.random.normal(
                    shape=spec.shape,
                    seed=None,
                ),
                dtype=spec.dtype,
            )
            for spec in self.input_specs.values()
        }

    def concrete_net(
        self, inputs: Dict[str, tf.Tensor | tf.TensorSpec] | None = None
    ) -> TFNetCallable:
        if inputs is None:
            inputs = self.input_specs
        return self.net.__call__.get_concrete_function(**inputs)

    def refine_weights(self) -> None:
        pass

    def run_eagerly(self, inputs: Dict[str, tf.Tensor]) -> Dict[str, tf.Tensor]:
        with EagerModeCtx(True), self.device:  # disable graph execution
            # TODO some op can only run on GPU (e.g. conv with NCHW)
            results = self.net(**inputs)
        return results

    @staticmethod
    def operators() -> List[Type[AbsOpBase]]:
        return list(ALL_TF_OPS)

    @staticmethod
    def add_seed_setter() -> None:
        register_seed_setter("tensorflow", tf.random.set_seed, overwrite=True)


class TFModelCPU(TFModel):
    @property
    def device(self) -> tf.device:
        return tf.device(tf.config.list_logical_devices("CPU")[0].name)


class TFModelCUDA(TFModel):
    @property
    def device(self) -> tf.device:
        gpus = tf.config.list_logical_devices("GPU")
        assert gpus, "No GPU available"
        return tf.device(gpus[0].name)



import os
import multiprocessing
import pickle
from pathlib import Path
from copy import deepcopy

import lz4.frame


def analyze_lcov(path):
    path = str(path)
    assert path.split(".")[-1] == "lz4"
    with lz4.frame.open(path, mode="rb") as fp:
        output_data = fp.read().decode("ISO-8859-1")  # LCOV issue.

    file_covs = output_data.split("end_of_record\n")[:-1]
    ret = {
        # 'filename': {
        #    'lines': [identifiers],
        #    'branches': [identifiers],
        #    'lf': # lines total,
        #    'bf': # branches total,
        # }
    }

    # SF: source file
    # sperated by end_of_record

    # FN: <line number>,<function name>

    for file_cov in file_covs:
        cov_lines = file_cov.split("\n")
        assert cov_lines[0].startswith("SF:")
        filename = cov_lines[0][3:]

        lines = set()
        branches = set()

        # LH: # line hits
        # FNH: # function hits | note we consider source-level functions. e.g., templates are one function.
        # BRH: # branch hits
        # NOTE: LH might (slightly) != len(lines). we only consider len(lines).

        n_line_total = 0  # LF: # lines
        n_branch_total = 0  # BRF: # branches

        lf = 0
        brf = 0

        for cov_line in cov_lines[1:]:
            cov_line = cov_line.rstrip("\n")
            if cov_line.startswith("DA:"):
                # DA: <line number>,<execution count> for each instrumented line
                line_number, exec_count = cov_line[3:].split(",")
                if exec_count != "0":
                    lines.add(int(line_number))
                n_line_total += 1
            elif cov_line.startswith("BRDA:"):
                # BRDA: <line number>,<block number>,<branch number>,<taken>
                line_number, block_number, branch_number, taken = cov_line[5:].split(
                    ","
                )
                if taken != "-" and taken != "0":
                    branches.add(line_number + ":" + block_number + ":" + branch_number)
                n_branch_total += 1
            # elif cov_line.startswith('LH:'):
            #     n_line_hit = int(cov_line[3:])
            elif cov_line.startswith("LF:"):
                lf = int(cov_line[3:])
            # elif cov_line.startswith('FNH:'):
            #     n_func_hit = int(cov_line[4:])
            elif cov_line.startswith("BRF:"):
                brf = int(cov_line[4:])
            else:
                pass
        # # BRF might be even smaller than branches you hit. we conservatively use n_branch_total.
        n_branch_total = max(brf, n_branch_total)
        # Similarly
        n_line_total = max(lf, n_line_total)

        assert (
            len(lines) <= n_line_total
        ), f"{len(lines)} <= {n_line_total} in {filename}"
        assert (
            len(branches) <= n_branch_total
        ), f"{len(branches)} <= {n_branch_total} in {filename}"

        ret[filename] = {
            "lines": lines,
            "branches": branches,
            "lf": n_line_total,
            "bf": n_branch_total,
        }

    return ret


def analyze_folder(folder, redo=False, max_time=None):
    return_write_name = os.path.join(folder, "merged_cov.pkl")

    if os.path.exists(return_write_name) and not redo:
        print("===> {} already exists.".format(return_write_name))
        with open(return_write_name, "rb") as fp:
            return pickle.load(fp)

    file_list = set()
    for file in Path(folder).rglob("*.lcov.lz4"):
        file_list.add(file)

    assert len(file_list) > 0, "no lcov files found in {}".format(folder)

    file_list = sorted(list(file_list), key=lambda x: int(str(x.name).split(".")[-3]))
    pool = multiprocessing.Pool(
        processes=min(len(file_list), multiprocessing.cpu_count())
    )
    results = pool.map(analyze_lcov, file_list)

    assert len(results) == len(file_list)
    stats = open(os.path.join(folder, "stats.csv"), "r").read().split("\n")
    if len(stats[-1]) == 0:
        stats = stats[:-1]
    assert len(stats) == len(results), f"{len(stats)} != {len(results)}"

    ret = {
        # 'time': time point. { # ! Not time duration.
        #   'n_model': # how many models got successfully executed?
        #   'merged_cov': coverage dictionary.
        # }
    }

    def merge_cov(current, rhs, hint=""):
        # {
        #     'lines': lines,
        #     'branches': branches,
        #     'lf': n_line_total,
        #     'bf': n_branch_total,
        # }

        for k in set(current.keys()).union(set(rhs.keys())):
            if k not in current:
                current[k] = rhs[k]
                continue
            elif k not in rhs:
                continue
            else:
                current[k]["lines"] = current[k]["lines"].union(rhs[k]["lines"])
                current[k]["branches"] = current[k]["branches"].union(
                    rhs[k]["branches"]
                )

                if current[k]["lf"] != rhs[k]["lf"]:
                    print(
                        f'[WARNING] total line {current[k]["lf"]} != {rhs[k]["lf"]} in {k} '
                        + hint
                    )
                    current[k]["lf"] = max(current[k]["lf"], rhs[k]["lf"])

                if current[k]["bf"] != rhs[k]["bf"]:
                    print(
                        f'[WARNING] total branch {current[k]["bf"]} != {rhs[k]["bf"]} in {k} '
                        + hint
                    )
                    current[k]["bf"] = max(current[k]["bf"], rhs[k]["bf"])

        return current

    current_cov = {}
    current_time = 0
    for i, r in enumerate(results):
        tokens = stats[i].split(",")
        t = tokens[0]
        n_model = tokens[2]
        current_time += float(t)
        if max_time is not None and current_time > max_time:
            break
        ret[current_time] = {}
        ret[current_time]["n_model"] = int(n_model)
        current_cov = merge_cov(current_cov, r, f"merging {file_list[i]}")
        ret[current_time]["merged_cov"] = deepcopy(current_cov)

    with open(return_write_name, "wb") as fp:
        pickle.dump(ret, fp)


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(
        description="Analyze coverage and plot visualizations."
    )
    parser.add_argument(
        "-f", "--folders", type=str, nargs="+", help="bug report folder"
    )
    parser.add_argument("--redo", action="store_true", help="redo analysis")
    parser.add_argument(
        "--max_time",
        type=int,
        default=60 * 60 * 4,
        help="max time in seconds for coverage evaluation",
    )
    args = parser.parse_args()

    for folder in args.folders:
        analyze_folder(folder, redo=args.redo, max_time=args.max_time)



from functools import partial
from typing import List, Type

import tensorflow as tf  # type: ignore
from keras import layers

from nnsmith.abstract.op import *
from nnsmith.materialize import framework_operator_impl
from nnsmith.materialize.tensorflow.dialect import *

# core dialect + some future PyTorch-only Operators.
TF_REALIZABLE_OPS = FULL_OPERATOR_SETS["core"] + FULL_OPERATOR_SETS["tensorflow"]
# TF_REALIZABLE_OPS = [NHWCConv2dSamePad, NHWCConv2dValidPad] # [Add, Dense, LocalRespNorm]
ALL_TF_OPS: List[Type[AbsOpBase]] = []

operator_impl = partial(framework_operator_impl, TF_REALIZABLE_OPS, ALL_TF_OPS)


"""Implement TensorFlow forward Callables for operator classes"""


@operator_impl(Constant)
def forward_fn(op: Constant):
    dtype = op.abs_tensor.dtype.tensorflow()
    data = tf.cast(tf.random.normal(op.abs_tensor.shape), dtype)
    return lambda: tf.constant(data, dtype=dtype)


@operator_impl(ReLU)
def forward_fn(op: ReLU):
    return tf.nn.relu


@operator_impl(GELU)
def forward_fn(op: GELU):
    return tf.nn.gelu


@operator_impl(LeakyReLU)
def forward_fn(op: LeakyReLU):
    return tf.nn.leaky_relu


@operator_impl(Sigmoid)
def forward_fn(op: Sigmoid):
    return tf.sigmoid


@operator_impl(Cos)
def forward_fn(op: Cos):
    return tf.cos


@operator_impl(Asin)
def forward_fn(op: Asin):
    return tf.asin


@operator_impl(Acos)
def forward_fn(op: Acos):
    return tf.acos


@operator_impl(Tan)
def forward_fn(op: Tan):
    return tf.tan


@operator_impl(Atan)
def forward_fn(op: Atan):
    return tf.atan


@operator_impl(Abs)
def forward_fn(op: Abs):
    return tf.abs


@operator_impl(Where)
def forward_fn(op: Where):
    return tf.where


@operator_impl(Add)
def forward_fn(op: Add):
    return tf.add


@operator_impl(Sub)
def forward_fn(op: Sub):
    return tf.math.subtract


@operator_impl(Mul)
def forward_fn(op: Mul):
    return tf.multiply


@operator_impl(Div)
def forward_fn(op: Div):
    return tf.divide


@operator_impl(Max)
def forward_fn(op: Max):
    return tf.maximum


@operator_impl(Min)
def forward_fn(op: Min):
    return tf.minimum


@operator_impl(Equal)
def forward_fn(op: Equal):
    return tf.equal


@operator_impl(Greater)
def forward_fn(op: Greater):
    return tf.greater


@operator_impl(Less)
def forward_fn(op: Less):
    return tf.less


@operator_impl(And)
def forward_fn(op: And):
    return tf.logical_and


@operator_impl(Or)
def forward_fn(op: Or):
    return tf.logical_or


@operator_impl(Xor)
def forward_fn(op: Xor):
    return tf.math.logical_xor


@operator_impl(Pow)
def forward_fn(op: Pow):
    return tf.pow


@operator_impl(Floor)
def forward_fn(op: Floor):
    return tf.floor


@operator_impl(Ceil)
def forward_fn(op: Ceil):
    return tf.math.ceil


@operator_impl(Clip)
def forward_fn(op: Clip):
    if op.input_like[0].dtype in DTYPE_GEN_FLOATS:
        return lambda x: tf.clip_by_value(x, -1.5, 1.5)
    else:
        return lambda x: tf.clip_by_value(x, -1, 1)


@operator_impl(Round)
def forward_fn(op: Round):
    return tf.round


@operator_impl(Sqrt)
def forward_fn(op: Sqrt):
    return tf.sqrt


@operator_impl(Log2)
def forward_fn(op: Log2):
    return tf.experimental.numpy.log2


@operator_impl(Neg)
def forward_fn(op: Neg):
    return tf.negative


@operator_impl(Softmax)
def forward_fn(op: Softmax):
    return lambda x: tf.nn.softmax(
        logits=tf.ensure_shape(x, op.input_like[0].shape),
        axis=op.dim,
    )


@operator_impl(Slice)
def forward_fn(op: Slice):
    reg = op.extra_attrs["region"]
    shape = op.input_like[0].shape
    dim_s = shape[op.extra_attrs["axis"]]
    start, end = op.start, op.end
    if reg in ["left", "mid"]:
        start -= dim_s
    # actual end would be 0, which is not really 'left'
    if reg == "left" and end < dim_s and end != Slice.INT_MAX:
        end -= dim_s
    idx = tuple(
        slice(None, None) if i != op.extra_attrs["axis"] else slice(start, end, op.step)
        for i in range(op.extra_attrs["ndims"])
    )

    return lambda x: x[idx]


@operator_impl(BatchNorm2d)
def forward_fn(op: BatchNorm2d):
    return layers.BatchNormalization(
        axis=1,
        dtype=op.input_like[0].dtype.tensorflow(),
        autocast=False,
    )  # NCHW


@operator_impl(Reshape)
def forward_fn(op: Reshape):
    def _reshape(x):
        return tf.reshape(x, op.target_shape)

    return _reshape


@operator_impl(Transpose)
def forward_fn(op: Transpose):
    def _transpose(x: tf.Tensor):
        aten = op.input_like[0]
        dim0, dim1 = op._init_swap_dims(aten.shape)
        perm = list(range(aten.ndims))
        perm[dim0], perm[dim1] = perm[dim1], perm[dim0]
        return tf.transpose(x, perm=perm)

    return _transpose


@operator_impl(Dense)
def forward_fn(op: Dense):
    return layers.Dense(
        units=op.ofeat, dtype=op.input_like[0].dtype.tensorflow(), autocast=False
    )


@operator_impl(LocalRespNorm)
def forward_fn(op: LocalRespNorm):
    def _lrn(x):
        return tf.raw_ops.LRN(
            input=x,
            depth_radius=op.depth_radius,
            bias=op.extra_attrs["bias"],
            alpha=op.extra_attrs["alpha"],
            beta=op.extra_attrs["beta"],
        )

    return _lrn


@operator_impl(NHWCConv2d)
def forward_fn(op: NHWCConv2d):
    return layers.Conv2D(
        filters=op.out_channels,
        kernel_size=(op.kernel_h_size, op.kernel_w_size),
        strides=(op.stride, op.stride),
        data_format="channels_last",
        dilation_rate=(op.dilation_h, op.dilation_w),
        padding=op.extra_attrs["padding"],
        dtype=op.input_like[0].dtype.tensorflow(),
        autocast=False,
    )


@operator_impl(Squeeze)
def forward_fn(op: Squeeze):
    if op.extra_attrs["reduce_dim"] is not None:
        return lambda x: tf.squeeze(x, axis=op.extra_attrs["reduce_dim"])
    return lambda x: tf.squeeze(x)


@operator_impl(ReduceSum)
def forward_fn(op: ReduceSum):
    if op.extra_attrs["reduce_dim"] is not None:
        return lambda x: tf.math.reduce_sum(x, axis=op.extra_attrs["reduce_dim"])
    return lambda x: tf.math.reduce_sum(x)


@operator_impl(ReduceMin)
def forward_fn(op: ReduceMin):
    if op.extra_attrs["reduce_dim"] is not None:
        return lambda x: tf.math.reduce_min(x, axis=op.extra_attrs["reduce_dim"])
    return lambda x: tf.math.reduce_min(x)


@operator_impl(ReduceMax)
def forward_fn(op: ReduceMax):
    if op.extra_attrs["reduce_dim"] is not None:
        return lambda x: tf.math.reduce_max(x, axis=op.extra_attrs["reduce_dim"])
    return lambda x: tf.math.reduce_max(x)


@operator_impl(ReduceMean)
def forward_fn(op: ReduceMean):
    if op.extra_attrs["reduce_dim"] is not None:
        return lambda x: tf.math.reduce_mean(x, axis=op.extra_attrs["reduce_dim"])
    return lambda x: tf.math.reduce_mean(x)


@operator_impl(ReduceProd)
def forward_fn(op: ReduceProd):
    if op.extra_attrs["reduce_dim"] is not None:
        return lambda x: tf.math.reduce_prod(x, axis=op.extra_attrs["reduce_dim"])
    return lambda x: tf.math.reduce_prod(x)


@operator_impl(ArgMin)
def forward_fn(op: ArgMin):
    if op.extra_attrs["reduce_dim"] is not None:
        return lambda x: tf.math.argmin(x, axis=op.extra_attrs["reduce_dim"])
    return lambda x: tf.math.argmin(x)


@operator_impl(ArgMin)
def forward_fn(op: ArgMin):
    if op.extra_attrs["reduce_dim"] is not None:
        return lambda x: tf.math.argmax(x, axis=op.extra_attrs["reduce_dim"])
    return lambda x: tf.math.argmax(x)


@operator_impl(Tril)
def forward_fn(op: Tril):
    return lambda x: tf.experimental.numpy.tril(x, k=op.diagonal)


@operator_impl(Triu)
def forward_fn(op: Triu):
    return lambda x: tf.experimental.numpy.triu(x, k=op.diagonal)


@operator_impl(Concat)
def forward_fn(op: Concat):
    axis = op.extra_attrs["axis"]
    return lambda *args: tf.concat(args, axis=axis)


@operator_impl(Cast)
def forward_fn(op: Cast):
    return lambda x: tf.cast(x, dtype=op.extra_attrs["to"].tensorflow())


@operator_impl(TFMatMul)
def forward_fn(op: TFMatMul):
    return tf.matmul



"""Just to figure out operators types and connections.
"""

from collections import Counter
import os
from multiprocessing import Pool, cpu_count
from typing import Dict, Set
import re
import pickle

import onnx
import pandas as pd

from tvm import relay

from nnsmith.backends import BackendFactory


def relay_op_cluster(mod, ignore_arg=False, verbose=False, use_counter=False):
    mod = relay.transform.InferType()(mod)
    op2type = {}

    def visit(node):
        def comment_remover(text):
            def replacer(match):
                s = match.group(0)
                if s.startswith("/"):
                    return " "  # note: a space and not an empty string
                else:
                    return s

            pattern = re.compile(
                r'//.*?$|/\*.*?\*/|\'(?:\\.|[^\\\'])*\'|"(?:\\.|[^\\"])*"',
                re.DOTALL | re.MULTILINE,
            )
            return re.sub(pattern, replacer, text)

        # the trick is: make a signature string. lmao.
        if isinstance(node, relay.Call):
            statement = comment_remover(str(node).splitlines()[-1]).replace(" ", "")
            num_args = len(node.type_args)
            attr_str = ",".join(statement[:-1].split(",")[num_args:])
            op_str = str(node.op)

            if attr_str == "":
                attr_str = None

            if ignore_arg:
                arg_type_str = None
            else:
                arg_type_str = str(node.type_args).replace(" ", "")

            hash_str = f"{arg_type_str}@{attr_str}"
            if verbose:
                print(f"[DEBUG] statement={statement}")
                print(f"[DEBUG] op_str={op_str}")
                print(f"[DEBUG] arg_type_str={arg_type_str}")
                print(f"[DEBUG] attr_str={attr_str}")
                print(f"[DEBUG] hash_str={hash_str}")

            if use_counter:
                op2type.setdefault(op_str, Counter()).update({hash_str: 1})
            else:
                op2type.setdefault(op_str, set()).update({hash_str: 1})

    for func in mod.functions.values():
        relay.analysis.post_order_visit(func, lambda node: visit(node))
    return op2type


def analyze_one_relay(model_path, use_counter=False) -> Dict[str, Set[str]]:
    """Return <op name> -> tag (a string)"""
    if "FAILURE" in model_path:
        return {}

    onnx_model = BackendFactory.get_onnx_proto(model_path)
    inp_spec, _ = BackendFactory.analyze_onnx_io(onnx_model)
    shape_dict = {name: inp_spec[name].shape for name in inp_spec}
    mod, _ = relay.frontend.from_onnx(onnx_model, shape_dict, freeze_params=True)

    return relay_op_cluster(mod)


def analyze_one(model_path):
    if "FAILURE" in model_path:
        return set(), set()

    model = onnx.load(model_path)
    output_to_op_t = {}

    nodes = set()
    edges = set()

    for node in model.graph.node:
        nodes.add(node.op_type)
        for o in node.output:
            output_to_op_t[o] = node.op_type

    for node in model.graph.node:
        for i in node.input:
            if i in output_to_op_t:
                edges.add((output_to_op_t[i], node.op_type))

    return nodes, edges


def analyze_folders(folders, cache_dir=None, force=False, n_limit=None):
    res = []

    __CACHE_FILE__ = "onnx_analysis_cache.pkl"
    for folder in folders:
        if (
            os.path.exists(os.path.join(cache_dir, __CACHE_FILE__))
            and not force
            and cache_dir is not None
        ):
            print(
                "===> {} already exists.".format(os.path.join(folder, __CACHE_FILE__))
            )
            with open(os.path.join(cache_dir, __CACHE_FILE__), "rb") as fp:
                return pickle.load(fp)

    times = []
    file_hubs = []
    least_time = None
    assert n_limit is None or len(n_limit) == len(folders)
    for i, folder in enumerate(folders):
        df = pd.read_csv(
            os.path.join(folder, "gentime.csv"), usecols=[0, 1], header=None
        )
        ts = df[0].to_numpy()

        times.append(ts)
        files = df[1].tolist()
        if n_limit is not None:
            files = files[: n_limit[i]]
        files = [os.path.join(folder, f) for f in files]
        file_hubs.append(files)

        time_cost = ts.sum()
        if least_time is None or time_cost < least_time:
            least_time = time_cost

    # only consider rows that < least_time
    for i, ts in enumerate(times):
        file_hubs[i] = file_hubs[i][: (ts < least_time).sum()]

    for files in file_hubs:
        nodes = set()
        edges = set()

        with Pool(min(cpu_count(), len(files))) as p:
            for n, e in p.imap_unordered(analyze_one, files):
                nodes |= n
                edges |= e

        res.append((nodes, edges))

    if cache_dir is not None:
        with open(os.path.join(cache_dir, __CACHE_FILE__), "wb") as fp:
            pickle.dump(res, fp)
    return res


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("--folders", type=str, nargs="+", required=True)
    parser.add_argument("--tags", type=str, nargs="+", default=None)
    # should compare models within same generation duration.
    parser.add_argument("--nlim", type=int, nargs="+", default=None)
    parser.add_argument("--output", type=str, default="results")
    parser.add_argument("--force", action="store_true")
    args = parser.parse_args()

    if args.tags is None:
        args.tags = [os.path.split(f)[-1].split("-")[0] for f in args.folders]
    else:
        assert len(args.tags) == len(args.folders)

    results = analyze_folders(
        args.folders, cache_dir=args.output, force=args.force, n_limit=args.nlim
    )
    for tag, (nodes, edges) in zip(args.tags, results):
        print(f"{tag}:\t nodes: {len(nodes)};\t edges: {len(edges)}")

    import matplotlib.pyplot as plt
    from matplotlib_venn import venn2, venn2_circles, venn3, venn3_circles

    node_list = []
    edge_list = []
    for i in range(len(results)):
        nodes, edges = results[i]
        node_list.append(nodes)
        edge_list.append(edges)

    if len(node_list) == 2:
        venn2(subsets=node_list, set_labels=[f"$\\bf{{{t}}}$" for t in args.tags])
        venn2_circles(subsets=node_list, linestyle="dashed")
    elif len(node_list) == 3:
        v = venn3(subsets=node_list, set_labels=[f"$\\bf{{{t}}}$" for t in args.tags])
        hatches = ["\\", ".", "*"]
        circles = ["MediumVioletRed", "SeaGreen", "Lavender"]
        for idx, id in enumerate(["100", "010", "001", "111"]):
            if v.get_label_by_id(id) is None:
                continue

            cnt = int(v.get_label_by_id(id).get_text())

            if id != "111":
                v.get_patch_by_id(id).set_alpha(0.5)
                v.get_patch_by_id(id).set_hatch(hatches[idx])
                v.get_patch_by_id(id).set_edgecolor(circles[idx])
                v.get_patch_by_id(id).set_linewidth(2)
                v.get_patch_by_id(id).set_linestyle("--")

    plt.title("Venn Diagram of Covered ONNX Operators")
    plt.savefig(
        f'{os.path.join(args.output, "onnx_node_venn")}.png', bbox_inches="tight"
    )
    plt.savefig(
        f'{os.path.join(args.output, "onnx_node_venn")}.pdf', bbox_inches="tight"
    )
    plt.close()

    if len(node_list) == 2:
        venn2(subsets=edge_list, set_labels=[f"$\\bf{{{t}}}$" for t in args.tags])
        venn2_circles(subsets=edge_list, linestyle="dashed")
    elif len(node_list) == 3:
        v = venn3(subsets=edge_list, set_labels=[f"$\\bf{{{t}}}$" for t in args.tags])
        hatches = ["\\", ".", "*"]
        circles = ["MediumVioletRed", "SeaGreen", "Lavender"]
        for idx, id in enumerate(["100", "010", "001", "111"]):
            if v.get_label_by_id(id) is None:
                continue

            cnt = int(v.get_label_by_id(id).get_text())

            if id != "111":
                v.get_patch_by_id(id).set_alpha(0.5)
                v.get_patch_by_id(id).set_hatch(hatches[idx])
                v.get_patch_by_id(id).set_edgecolor(circles[idx])
                v.get_patch_by_id(id).set_linewidth(2)
                v.get_patch_by_id(id).set_linestyle("--")
    plt.title("Venn Diagram of Covered ONNX Operators Edges")
    plt.savefig(
        f'{os.path.join(args.output, "onnx_edge_venn")}.png', bbox_inches="tight"
    )
    plt.savefig(
        f'{os.path.join(args.output, "onnx_edge_venn")}.pdf', bbox_inches="tight"
    )
    plt.close()



import multiprocessing as mp
import sys
import traceback
from abc import ABC, abstractmethod
from typing import Callable, Dict, List, Optional, Union

import numpy as np

from nnsmith.abstract.dtype import DType
from nnsmith.abstract.tensor import AbsTensor
from nnsmith.difftest import assert_allclose
from nnsmith.error import InternalError
from nnsmith.logging import CORE_LOG
from nnsmith.materialize import BugReport, Model, Oracle, Stage, Symptom, TestCase

BackendCallable = Callable[[Dict[str, np.ndarray]], Dict[str, np.ndarray]]


class BackendFactory(ABC):
    def __init__(self, target="cpu", optmax: bool = False):
        super().__init__()
        self.target = target
        self.optmax = optmax

    @property
    @abstractmethod
    def system_name(self) -> str:
        pass

    @property
    def version(self) -> str:
        return "unknown"

    def __str__(self) -> str:
        return f"{self.system_name} ({self.target} opt: {self.optmax})"

    @staticmethod
    def make_random_input(
        input_like: Dict[str, AbsTensor], low=1, high=2
    ) -> Dict[str, np.ndarray]:
        return {
            name: np.random.uniform(low=low, high=high, size=aten.shape).astype(
                aten.dtype.numpy()
            )
            for name, aten in input_like.items()
        }

    @classmethod
    def skip_dtypes(cls) -> List[DType]:
        return []

    @abstractmethod
    def make_backend(self, model: Model) -> BackendCallable:
        raise NotImplementedError

    def critical_assert_dispatchable(self, model: Model):
        if not self.make_backend.dispatch(type(model)):
            CORE_LOG.critical(
                f"[Not implemented] {type(self).__name__} for {type(model).__name__}!\n"
                "Check https://github.com/ise-uiuc/nnsmith#backend-model-support for compatile `model.type` and `backend.type`."
            )
            sys.exit(1)

    def checked_make_backend(self, model: Model) -> BackendCallable:
        self.critical_assert_dispatchable(model)
        return self.make_backend(model)

    def checked_compile(self, testcase: TestCase) -> Union[BackendCallable, BugReport]:
        try:  # compilation
            return self.checked_make_backend(testcase.model)
        except InternalError as e:
            raise e
        except Exception:
            return BugReport(
                testcase=testcase,
                system=self.system_name,
                symptom=Symptom.EXCEPTION,
                stage=Stage.COMPILATION,
                log=traceback.format_exc(),
                version=self.version,
            )

    def checked_exec(
        self, executable: BackendCallable, testcase: TestCase
    ) -> Union[Dict[str, np.ndarray], BugReport]:
        input = None if testcase.oracle is None else testcase.oracle.input
        if input is None:
            input = self.make_random_input(testcase.model.input_like)
            testcase = TestCase(
                model=testcase.model, oracle=Oracle(input=input, output=None)
            )

        try:  # execution
            return executable(input)
        except InternalError as e:
            raise e
        except Exception:
            return BugReport(
                testcase=testcase,
                system=self.system_name,
                symptom=Symptom.EXCEPTION,
                stage=Stage.EXECUTION,
                log=traceback.format_exc(),
                version=self.version,
            )

    def checked_compile_and_exec(
        self, testcase: TestCase, crash_safe=False, timeout=None
    ) -> Union[Dict[str, np.ndarray], BugReport]:
        # pre-check model dispatchability
        self.critical_assert_dispatchable(testcase.model)
        if (
            not crash_safe and timeout is None
        ):  # not crash safe, compile & exec natively in current process.
            bug_or_exec = self.checked_compile(testcase)
            if isinstance(bug_or_exec, BugReport):
                return bug_or_exec
            return self.checked_exec(bug_or_exec, testcase)
        else:  # crash safe, compile & exec in a separate process.
            if timeout is not None:
                assert isinstance(
                    timeout, int
                ), "timeout are `seconds` => must be an integer."

        # TODO: optimize to shared memory in the future (Python 3.8+)
        # https://docs.python.org/3/library/multiprocessing.shared_memory.html
        # NOTE: Similar implementation as Tzer.
        with mp.Manager() as manager:
            shared_dict = manager.dict(
                {
                    "symptom": None,
                    "stage": Stage.COMPILATION,
                    "log": None,
                    "output": None,
                    "uncaught_exception": None,
                }
            )

            def crash_safe_compile_exec(sdict):
                try:
                    CORE_LOG.debug(f"[FORK] Compiling.")
                    bug_or_exec = self.checked_compile(testcase)
                    if isinstance(bug_or_exec, BugReport):
                        sdict["symptom"] = bug_or_exec.symptom
                        sdict["log"] = bug_or_exec.log
                        return

                    CORE_LOG.debug(f"[FORK] Executing.")
                    sdict["stage"] = Stage.EXECUTION
                    bug_or_result = self.checked_exec(bug_or_exec, testcase)
                    CORE_LOG.debug(f"[FORK] Done.")
                    if isinstance(bug_or_result, BugReport):
                        sdict["symptom"] = bug_or_result.symptom
                        sdict["log"] = bug_or_result.log
                        return

                    sdict["output"] = bug_or_result
                except Exception as e:
                    sdict["uncaught_exception"] = e

            p = mp.Process(target=crash_safe_compile_exec, args=(shared_dict,))

            p.start()
            p.join(timeout=timeout)
            if p.is_alive():
                p.terminate()
                p.join()
                assert not p.is_alive()
                return BugReport(
                    testcase=testcase,
                    system=self.system_name,
                    symptom=Symptom.TIMEOUT,
                    stage=shared_dict["stage"],
                    log=f"Timeout after {timeout} seconds.",
                    version=self.version,
                )

            if shared_dict["output"] is not None:
                return shared_dict["output"]

            if shared_dict["uncaught_exception"] is not None:
                CORE_LOG.critical(
                    f"Found uncaught {type(shared_dict['uncaught_exception'])} in crash safe mode."
                )
                raise shared_dict["uncaught_exception"]

            if p.exitcode != 0:
                return BugReport(
                    testcase=testcase,
                    system=self.system_name,
                    symptom=Symptom.SEGFAULT,
                    stage=shared_dict["stage"],
                    log=f"Process crashed with exit code: {p.exitcode}",
                    version=self.version,
                )
            else:
                return BugReport(
                    testcase=testcase,
                    system=self.system_name,
                    symptom=shared_dict["symptom"],
                    stage=shared_dict["stage"],
                    log=shared_dict["log"],
                    version=self.version,
                )

    def verify_results(
        self, output: Dict[str, np.ndarray], testcase: TestCase, equal_nan=True
    ) -> Optional[BugReport]:
        try:  # verification
            assert_allclose(
                output,
                testcase.oracle.output,
                self.__str__(),
                testcase.oracle.provider,
                equal_nan=equal_nan,
            )
        except AssertionError:
            return BugReport(
                testcase=testcase,
                system=self.system_name,
                symptom=Symptom.INCONSISTENCY,
                stage=Stage.VERIFICATION,
                log=traceback.format_exc(),
                version=self.version,
            )
        except Exception:
            return BugReport(
                testcase=testcase,
                system=self.system_name,
                symptom=Symptom.EXCEPTION,
                stage=Stage.VERIFICATION,
                log=traceback.format_exc(),
                version=self.version,
            )

    def verify_testcase(
        self, testcase: TestCase, equal_nan=True
    ) -> Optional[BugReport]:
        executable = self.checked_compile(testcase)
        if isinstance(executable, BugReport):
            return executable

        output = self.checked_exec(executable, testcase)
        if isinstance(output, BugReport):
            return output

        if testcase.oracle is not None and testcase.oracle.output is not None:
            return self.verify_results(output, testcase, equal_nan=equal_nan)

        return None

    def make_testcase(
        self,
        model: Model,
        input: Dict[str, np.ndarray] = None,
        crash_safe=False,
        timeout=None,
    ) -> Union[BugReport, TestCase]:
        if input is None:
            input = self.make_random_input(model.input_like)

        partial_testcase = TestCase(
            model=model, oracle=Oracle(input=input, output=None)
        )
        bug_or_res = self.checked_compile_and_exec(
            partial_testcase, crash_safe=crash_safe, timeout=timeout
        )
        if isinstance(bug_or_res, BugReport):
            return bug_or_res
        else:
            partial_testcase.oracle.output = bug_or_res

        try:  # compilation
            executable = self.checked_make_backend(model)
        except InternalError as e:
            raise e
        except Exception:
            return BugReport(
                TestCase(model=model, oracle=None),  # None means no oracle
                system=self.system_name,
                symptom=Symptom.EXCEPTION,
                stage=Stage.COMPILATION,
                log=traceback.format_exc(),
                version=self.version,
            )

        if not input:
            # generate inputs automatically.
            input = self.make_random_input(model.input_like)

        try:  # execution
            output = executable(input)
        except InternalError as e:
            raise e
        except Exception:
            return BugReport(
                TestCase(model, Oracle(input=input, output=None)),
                system=self.system_name,
                symptom=Symptom.EXCEPTION,
                stage=Stage.EXECUTION,
                log=traceback.format_exc(),
                version=self.version,
            )

        return TestCase(
            model, Oracle(input=input, output=output, provider=self.system_name)
        )

    @staticmethod
    def init(name, target="cpu", optmax=True, **kwargs):
        if name is None:
            raise ValueError(
                "Backend type cannot be None. Specify via `backend.type=[onnxruntime | tvm | tensorrt | tflite | xla ]`"
            )

        if target == "gpu":
            target = "cuda"  # `gpu` means `cuda` by default.

        if name == "onnxruntime":
            from nnsmith.backends.onnxruntime import ORTFactory

            return ORTFactory(
                target=target,
                optmax=optmax,
                **kwargs,
            )
        elif name == "tvm":
            from nnsmith.backends.tvm import TVMFactory

            # default executor is graph
            kwargs["executor"] = kwargs.get("executor", "graph")
            return TVMFactory(
                target=target,
                optmax=optmax,
                **kwargs,
            )
        elif name == "tensorrt":
            from nnsmith.backends.tensorrt import TRTFactory

            return TRTFactory(
                target=target,
                optmax=optmax,
                **kwargs,
            )
        elif name == "tflite":
            from nnsmith.backends.tflite import TFLiteFactory

            return TFLiteFactory(
                target=target,
                optmax=optmax,
                **kwargs,
            )
        elif name == "xla":
            from nnsmith.backends.xla import XLAFactory

            return XLAFactory(
                target=target,
                optmax=optmax,
                **kwargs,
            )
        elif name == "torchjit":
            from nnsmith.backends.torchjit import TorchJITFactory

            return TorchJITFactory(target=target, optmax=optmax, **kwargs)
        else:
            raise ValueError(f"unknown backend: {name}")



import logging

VIZ_LOG = logging.getLogger("viz")
FUZZ_LOG = logging.getLogger("fuzz")
MGEN_LOG = logging.getLogger("mgen")
SMT_LOG = logging.getLogger("smt")
EXEC_LOG = logging.getLogger("exec")
DTEST_LOG = logging.getLogger("dtest")
CORE_LOG = logging.getLogger("core")

TF_LOG = logging.getLogger("gen|tf")
TORCH_LOG = logging.getLogger("gen|torch")



import os
import random
from abc import ABC, abstractmethod
from copy import deepcopy
from functools import reduce
from inspect import signature
from typing import Dict, List, Optional, Tuple, Type, Union

import z3

from nnsmith.abstract.arith import *
from nnsmith.abstract.dtype import (
    DTYPE_GEN_ALL,
    DTYPE_GEN_COMPLEX,
    DTYPE_GEN_FLOATS,
    DTYPE_GEN_INTS,
    DTYPE_GEN_NON_BOOL,
    DType,
)
from nnsmith.abstract.extension import ACTIVATED_PATCH
from nnsmith.abstract.tensor import AbsTensor
from nnsmith.error import ConstraintCheck, SanityCheck

# There are following types of constraints at this point:
# 1. Shape variables must be greater than 0;
# 2. Shape variables must avoid devision by 0;
# 3. Extra constraints introduced by individual operators;


__MIN_RANK__ = 0
__MAX_RANK__ = 5

FULL_OPERATOR_SETS: Dict[str, List[Type["AbsOpBase"]]] = dict()

_PRAGMA_ONCE_CORE_OP = True  # have to use this dirty hack to avoid namespace violation.


class mark_abstract:
    def __init__(self, dialect):
        assert (
            dialect != "core" or _PRAGMA_ONCE_CORE_OP
        ), "`core` is exclusive to nnsmith.abstract.op. Please use other dilect names"
        self.dialect = dialect

    def __call__(self, op_type: Type["AbsOpBase"]) -> Type["AbsOpBase"]:
        op_type.dialect = self.dialect
        return op_type


class mark_materialize:
    def __init__(self, dialect: str, limit_domain=False):
        self.limit_domain = limit_domain
        self.dialect = dialect

    def __call__(self, op_type: Type["AbsOpBase"]) -> Type["AbsOpBase"]:
        op_list = FULL_OPERATOR_SETS.setdefault(self.dialect, [])

        if op_type not in op_list:
            op_type = mark_abstract(self.dialect)(op_type)
            op_type.limit_domain = self.limit_domain
            op_list.append(op_type)
            op_list.sort(key=lambda x: x.__name__)

        return op_type


def rank_from(start):
    return tuple(range(start, __MAX_RANK__ + 1))


def rank_range(start, end):
    return tuple(range(start, end + 1))


def rank_until(end):
    return tuple(range(__MIN_RANK__, end + 1))


def rank_all():
    return tuple(range(__MIN_RANK__, __MAX_RANK__ + 1))


def check_shape_fn(func):
    def wrapper_check_shape_fn(self, input_shapes):
        SanityCheck.true(
            self.out_ranks,
            "Empty output dimensions in {}".format(self.__class__.__name__),
        )
        SanityCheck.eq(
            len(input_shapes),
            self.n_input(),
            "{} requires {} inputs, but got {}".format(
                self.__class__.__name__, self.n_input(), len(input_shapes)
            ),
        )
        res = func(self, [s.deepcopy() for s in input_shapes])
        SanityCheck.eq(
            len(res),
            self.n_output(),
            "{} requires {} outputs, but got {}".format(
                self.__class__.__name__, self.n_output(), len(res)
            ),
        )
        return res

    return wrapper_check_shape_fn


def check_require_fn(func):
    def wrapper_check_require_fn(self, input_shapes: List[AbsTensor]):
        SanityCheck.eq(
            len(input_shapes),
            self.n_input(),
            "{} requires {} inputs, but got {}".format(
                self.__class__.__name__, self.n_input(), len(input_shapes)
            ),
        )
        return func(self, [s.deepcopy() for s in input_shapes])

    return wrapper_check_require_fn


def _prepend_to(x, max_dim):
    return [1 for _ in range(max_dim - len(x))] + x


def z3_bcast(
    x: Union[int, z3.ExprRef], y: Union[int, z3.ExprRef], *args: Union[int, z3.ExprRef]
):
    x, y = align_bvs(x, y)
    return (
        z3.simplify(z3.If(nnsmith_eq(y, 1), x, y))
        if len(args) == 0
        else z3_bcast(z3_bcast(x, y), *args)
    )


def int_bcast(x: int, *args: int):
    return x if (x != 1 or len(args) == 0) else int_bcast(*args)


def broadcast_shapes(
    *shapes: List[Union[z3.ExprRef, int]]
) -> List[Union[z3.ExprRef, int]]:
    """this function does not check the validity of broadcast. Please always pair it with broadcast_cons"""
    SanityCheck.gt(len(shapes), 0)
    if len(shapes) == 1:
        return shapes[0]
    max_dim = max(map(lambda x: len(x), shapes))
    out_shape = [None] * max_dim
    for j in range(max_dim):
        i = -j - 1
        args_dim_sz = [_prepend_to(x, max_dim)[i] for x in shapes]
        if any(isinstance(s, z3.ExprRef) for s in args_dim_sz):
            out_shape[i] = z3.simplify(z3_bcast(*args_dim_sz))
        else:
            out_shape[i] = int_bcast(*args_dim_sz)
    return out_shape


def broadcast_cons(*shapes: List[Union[z3.ExprRef, int]]) -> List[z3.BoolRef]:
    tgt_shape = broadcast_shapes(*shapes)
    cons = []
    max_dim = len(tgt_shape)
    for j in range(max_dim):
        i = -j - 1
        if isinstance(tgt_shape[i], z3.ExprRef):
            axis_cons = []
            for x in shapes:
                if len(x) > j:
                    axis_cons.append(
                        z3.Or(nnsmith_eq(x[i], tgt_shape[i]), nnsmith_eq(x[i], 1))
                    )
            axis_cons = z3.simplify(z3.And(*axis_cons))
            cons.append(axis_cons)
        else:
            args_dim_sz = [_prepend_to(x, max_dim)[i] for x in shapes]
            valid = all(
                nnsmith_eq(s, tgt_shape[i]) or nnsmith_eq(s, 1) for s in args_dim_sz
            )
            # TODO(JK): enable this after fixing issue #2
            # assert valid, "Invalid broadcast shapes {}. Specific dim sizes: {}".format(shapes, args_dim_sz)
            cons.append(z3.BoolVal(valid))
    return cons


def broadcast_cons_binary(*shapes: List[Union[z3.ExprRef, int]]) -> List[z3.BoolRef]:
    SanityCheck.eq(len(shapes), 2)
    tgt_shape = broadcast_shapes(*shapes)
    cons = []
    max_dim = len(tgt_shape)
    lhs, rhs = shapes
    lhs = _prepend_to(lhs, max_dim)
    rhs = _prepend_to(rhs, max_dim)
    for j in range(max_dim):
        i = -j - 1
        if isinstance(tgt_shape[i], z3.ExprRef):
            cons.append(
                z3.simplify(
                    z3.Or(
                        nnsmith_eq(lhs[i], 1),
                        nnsmith_eq(rhs[i], 1),
                        nnsmith_eq(lhs[i], rhs[i]),
                    )
                )
            )
        else:
            valid = (
                nnsmith_eq(lhs[i], 1)
                or nnsmith_eq(rhs[i], 1)
                or nnsmith_eq(lhs[i], rhs[i])
            )
            # TODO(JK): enable this after fixing issue #2
            # assert valid, "Invalid broadcast shapes lhs={}, rhs={}".format(lhs, rhs)
            cons.append(z3.BoolVal(valid))
    return cons


def broadcast_to_cons(*shapes: List[Union[z3.ExprRef, int]]) -> List[z3.BoolRef]:
    """Unidirectional broadcast. Last input is the target shape.

    Examples of valid unidirectional broadcast:
    [1, 2, 3] -> [0, 1, 2, 3]
    [1] -> [3]

    Examples of invalid unidirectional broadcast:
    [0, 1, 2, 3] -> [1, 2, 3]
    [3] -> [1]

    Logic: for each dim: src_dim == tgt_dim or src_dim == 1
    """
    srcs, tgt = shapes[:-1], shapes[-1]
    cons = []
    max_dim = len(tgt)
    for src in srcs:
        ConstraintCheck.true(len(src) <= max_dim)
        src = _prepend_to(src, max_dim)
        for i in range(max_dim):
            if isinstance(tgt[i], z3.ExprRef) or isinstance(src[i], z3.ExprRef):
                cons.append(
                    z3.simplify(
                        z3.Or(nnsmith_eq(src[i], 1), nnsmith_eq(src[i], tgt[i]))
                    )
                )
            else:
                valid = nnsmith_eq(src[i], 1) or nnsmith_eq(src[i], tgt[i])
                # TODO(JK): enable this after fixing issue #2
                # assert valid, "Invalid broadcast shapes lhs={}, rhs={}".format(lhs, rhs)
                cons.append(z3.BoolVal(valid))
    return cons


@mark_abstract("core")
class AbsOpBase(ABC):
    # number of parameters; None means it's fixed that can be inferred through `signature`.
    num_var_param = None
    # Require the input dimension sizes to be equivalent.
    same_inp_dims = False
    # input dtypes: enumerates all possible input dtype combinations. Size of the list is the number of combinations.
    # Each element is a tuple of allowed input dtypes. NOTE: len(list) can >= the # of inputs, for handling ops with arbitrary arity.
    # For example, [(DType.float32, DType.float32), (DType.float64, DType.float64), (DType.int32, DType.int32)] means that
    # this op can accept one of float32xfloat32, float64xfloat64, and int32xint32 as input dtypes.
    in_dtypes: List[Tuple[DType, ...]] = None  # Overwrite me!
    out_dtypes: List[Tuple[DType, ...]] = None

    limit_domain = False

    dialect = None

    def __init__(self):
        assert self.dialect, "Set dialect with `mark_dialect` or `mark_realize`"

        # `[3, 3]` this means this op requires 2 inputs. Where the 1st one has 2 dimensions, and the 2nd one has 3 dimensions.
        # `-1` means arbitrary dimantions; NOTE: but should be concretized during execution.
        # All symbols of correponding operator must be the constructor's parameters.
        # [ <inp0>(support_dim0, support_dim1, ...), <inp1>(...), ... ]
        self.inp_ranks = []
        # NOTE: the concrete values of out_ranks are not useful. Just make sure the length is correct.
        # NOTE: the output shape of input dimensions should be concretized during the execution.
        self.out_ranks = []
        # NOTE: the input of operator constructors are all Union[int, z3.ExprRef].
        self.extra_attrs = {}
        self._input_like = [None] * self.n_input()
        self._output_like = [None] * self.n_output()

    @classmethod
    def n_input(cls):
        assert cls.in_dtypes is not None
        return len(cls.in_dtypes[0]) if len(cls.in_dtypes) > 0 else 0

    @classmethod
    def n_output(cls):
        assert cls.out_dtypes is not None
        return len(cls.out_dtypes[0]) if len(cls.out_dtypes) > 0 else 0

    @classmethod
    def get_num_var_param(cls):
        if cls.num_var_param is None:
            return len(signature(cls.__init__).parameters) - 1
        return random.choice(cls.num_var_param)

    def bind_input_like(self, input_like: List[AbsTensor]):
        self._input_like = input_like

    def bind_output_like(self, output_like: List[AbsTensor]):
        self._output_like = output_like

    @property
    def input_like(self):
        return self._input_like

    @property
    def output_like(self):
        return self._output_like

    @abstractmethod  # Overload me!
    # Exception means rejection.
    def type_transfer(self, input_shapes: List[AbsTensor]) -> List[AbsTensor]:
        raise NotImplementedError

    @check_shape_fn  # Public API.
    def checked_type_transfer(self, input_shapes: List[AbsTensor]) -> List[AbsTensor]:
        return self.type_transfer(input_shapes)

    # Overload me!
    # Extra constraints for the input tensors.
    # Exception means rejection.
    def requires(self, input_shapes: List[AbsTensor]) -> List[Union[z3.ExprRef, bool]]:
        return []

    @abstractmethod
    def deduct_inp_ranks_and_dtype(
        self, out_abs_tensor: List[AbsTensor]
    ) -> List[Tuple[int, DType]]:
        raise NotImplementedError

    @check_require_fn  # Public API.
    def checked_requires(self, input_shapes):
        extra = []
        for f in ACTIVATED_PATCH.get(self.name(), []):
            extra.extend(f(self, input_shapes))
        return self.requires(input_shapes) + extra

    def n_floats(self, input_shapes: List[AbsTensor]) -> z3.ExprRef:
        return reduce(nnsmith_add, [i.nelement() for i in self.output_like])

    def flops(self, input_shapes):
        return 0

    def __repr__(self) -> str:
        return self.__class__.__name__

    @classmethod
    def __str__(cls) -> str:
        return cls.name()

    @classmethod
    def name(cls) -> str:
        if hasattr(cls, "dialect"):
            return cls.dialect + "." + cls.__name__.split(".")[-1]
        return cls.__name__.split(".")[-1]


class UnaryOpBase(AbsOpBase):
    in_dtypes = [(i,) for i in DTYPE_GEN_FLOATS]
    out_dtypes = [(i,) for i in DTYPE_GEN_FLOATS]

    def __init__(self):
        super().__init__()
        self.out_ranks = [rank_all()]


class BinaryOpBase(AbsOpBase):
    in_dtypes = [(i, i) for i in DTYPE_GEN_FLOATS]
    out_dtypes = [(i,) for i in DTYPE_GEN_FLOATS]

    def __init__(self):
        super().__init__()
        self.out_ranks = [rank_all()]


class TernaryOpBase(AbsOpBase):
    in_dtypes = [(i, i, i) for i in DTYPE_GEN_FLOATS]
    out_dtypes = [(i,) for i in DTYPE_GEN_FLOATS]

    def __init__(self):
        super().__init__()
        self.out_ranks = [rank_all()]


class ElementWiseUnaryOp(UnaryOpBase):
    def __init__(self):
        super().__init__()
        self.inp_ranks = [rank_all()]
        self.out_ranks = [rank_all()]

    def type_transfer(self, input_shapes: List[AbsTensor]) -> List[AbsTensor]:
        SanityCheck.eq(len(input_shapes), 1)
        return [input_shapes[0]]

    def deduct_inp_ranks_and_dtype(
        self, out_abs_tensor: List[AbsTensor]
    ) -> List[Tuple[int, DType]]:
        return [
            (out_abs_tensor[0].ndims, out_abs_tensor[0].dtype),
        ]


def bcast_rand_ndims(num_svars, target_ndims):
    res = [random.randint(0, target_ndims) for _ in range(num_svars)]
    res[random.randint(0, num_svars - 1)] = target_ndims
    return res


class BcastBinaryOp(BinaryOpBase):
    # by default, output dtype is the same as the first input dtype
    _bcast_out_dtypes = None

    def __init__(self):
        super().__init__()
        self.inp_ranks = [rank_all(), rank_all()]

    def type_transfer(self, input_shapes: List[AbsTensor]) -> List[AbsTensor]:
        tgt_shape = broadcast_shapes(*(ish.shape for ish in input_shapes))
        dtype = (
            input_shapes[0].dtype
            if self._bcast_out_dtypes is None
            else self._bcast_out_dtypes[0]
        )
        return [AbsTensor(tgt_shape, dtype)]

    def requires(self, input_shapes):
        return broadcast_cons_binary(*(ish.shape for ish in input_shapes))

    def deduct_inp_ranks_and_dtype(
        self, out_abs_tensor: List[AbsTensor]
    ) -> List[Tuple[int, DType]]:
        x, y = bcast_rand_ndims(2, out_abs_tensor[0].ndims)
        return [
            (x, out_abs_tensor[0].dtype),
            (y, out_abs_tensor[0].dtype),
        ]


class BcastBinaryOp1(BcastBinaryOp):  # +-*/ max min
    in_dtypes = [(i, i) for i in DTYPE_GEN_NON_BOOL]
    out_dtypes = [(i,) for i in DTYPE_GEN_NON_BOOL]
    _bcast_out_dtypes = None


class Comparator(BcastBinaryOp):  # > < =
    in_dtypes = [(i, i) for i in DTYPE_GEN_ALL]
    out_dtypes = [(DType.bool,)]
    _bcast_out_dtypes = [DType.bool]

    def deduct_inp_ranks_and_dtype(
        self, out_abs_tensor: List[AbsTensor]
    ) -> List[Tuple[int, DType]]:
        x, y = bcast_rand_ndims(2, out_abs_tensor[0].ndims)
        in_dtypes = random.choice(self.in_dtypes)
        return [
            (x, in_dtypes[0]),
            (y, in_dtypes[1]),
        ]


class IntegerComparator(Comparator):  # > < = for integer
    in_dtypes = [(i, i) for i in DTYPE_GEN_INTS]


class Logical(Comparator):  # logical and or xor
    in_dtypes = [(DType.bool, DType.bool)]


@mark_materialize("core")
class Where(TernaryOpBase):
    in_dtypes = [(DType.bool, i, i) for i in DTYPE_GEN_NON_BOOL]
    out_dtypes = [(i,) for i in DTYPE_GEN_NON_BOOL]

    def __init__(self):
        super().__init__()
        self.inp_ranks = [rank_all(), rank_all(), rank_all()]
        self.same_inp_dtypes = True

    def type_transfer(self, input_shapes: List[AbsTensor]) -> List[AbsTensor]:
        # assert len(input_shapes[0].shape) == len(input_shapes[1].shape)
        tgt_shape = broadcast_shapes(*(ish.shape for ish in input_shapes))
        dtype = input_shapes[1].dtype
        return [AbsTensor(tgt_shape, dtype)]

    def requires(self, input_shapes):
        return broadcast_cons(*(ish.shape for ish in input_shapes)) + [
            z3.BoolVal(input_shapes[1].dtype == input_shapes[2].dtype)
        ]

    def deduct_inp_ranks_and_dtype(
        self, out_abs_tensor: List[AbsTensor]
    ) -> List[Tuple[int, DType]]:
        x, y, z = bcast_rand_ndims(3, out_abs_tensor[0].ndims)
        return [
            (x, DType.bool),
            (y, out_abs_tensor[0].dtype),
            (z, out_abs_tensor[0].dtype),
        ]


# bcast binary ops from https://github.com/onnx/onnx/blob/master/docs/Broadcasting.md
Add: Type[BcastBinaryOp1] = mark_materialize("core")(
    type(
        "Add",
        (BcastBinaryOp1,),
        {"__module__": __name__},
    )
)
Sub: Type[BcastBinaryOp1] = mark_materialize("core")(
    type(
        "Sub",
        (BcastBinaryOp1,),
        {"__module__": __name__},
    )
)
Mul: Type[BcastBinaryOp1] = mark_materialize("core")(
    type(
        "Mul",
        (BcastBinaryOp1,),
        {"__module__": __name__},
    )
)
Max: Type[BcastBinaryOp1] = mark_materialize("core")(
    type(
        "Max",
        (BcastBinaryOp1,),
        {"__module__": __name__},
    )
)
Min: Type[BcastBinaryOp1] = mark_materialize("core")(
    type(
        "Min",
        (BcastBinaryOp1,),
        {"__module__": __name__},
    )
)

# Set to `IntegerComparator` as "float == float" easily leads to false positive.
Equal: Type[IntegerComparator] = mark_materialize("core")(
    type("Equal", (IntegerComparator,), {"__module__": __name__})
)
Greater: Type[Comparator] = mark_materialize("core")(
    type(
        "Greater",
        (Comparator,),
        {"__module__": __name__},
    )
)
Less: Type[Comparator] = mark_materialize("core")(
    type("Less", (Comparator,), {"__module__": __name__})
)
And: Type[Logical] = mark_materialize("core")(
    type(
        "And",
        (Logical,),
        {"__module__": __name__},
    )
)
Or: Type[Logical] = mark_materialize("core")(
    type(
        "Or",
        (Logical,),
        {"__module__": __name__},
    )
)
Xor: Type[Logical] = mark_materialize("core")(
    type(
        "Xor",
        (Logical,),
        {"__module__": __name__},
    )
)


class Input(AbsOpBase):
    in_dtypes = [()]
    out_dtypes = [(i,) for i in DTYPE_GEN_ALL]

    def __init__(self, dim: int):
        super().__init__()
        self.inp_ranks = []
        self.out_ranks = [(dim,)]
        self.abs_tensor: AbsTensor = None

    def type_transfer(self, input_shapes: List[AbsTensor]) -> List[AbsTensor]:
        SanityCheck.eq(len(input_shapes), 0)
        return [self.abs_tensor]

    def requires(self, input_shapes: List[AbsTensor]) -> List[Union[z3.BoolRef, bool]]:
        SanityCheck.eq(len(input_shapes), 0)
        return []

    def __str__(self):
        return "Input"

    def deduct_inp_ranks_and_dtype(
        self, out_abs_tensor: List[AbsTensor]
    ) -> List[Tuple[int, DType]]:
        pass

    @property
    def input_like(self):
        return []

    @property
    def output_like(self):
        return [self.abs_tensor]


class Constant(AbsOpBase):
    in_dtypes = [()]
    out_dtypes = [(i,) for i in DTYPE_GEN_ALL]

    def __str__(self) -> str:
        return self.name() + " " + str(self.extra_attrs).replace(":", "=")

    def __init__(self, dim: int):
        super().__init__()
        self.dim = dim
        self.inp_ranks = []
        self.out_ranks = [(dim,)]
        self.abs_tensor: AbsTensor = None

    def type_transfer(self, input_shapes: List[AbsTensor]) -> List[AbsTensor]:
        SanityCheck.eq(len(input_shapes), 0)
        return [self.abs_tensor]

    def requires(self, input_shapes: List[AbsTensor]) -> List[Union[z3.BoolRef, bool]]:
        SanityCheck.eq(len(input_shapes), 0)
        return []

    def __str__(self):
        return "Constant"

    def deduct_inp_ranks_and_dtype(
        self, out_abs_tensor: List[AbsTensor]
    ) -> List[Tuple[int, DType]]:
        pass

    @property
    def input_like(self):
        return []

    @property
    def output_like(self):
        return [self.abs_tensor]


class Placeholder:
    def __init__(self, ttype: AbsTensor):
        self.ttype = ttype
        self.inp_ranks = []
        self.out_ranks = [(ttype.ndims,)]
        self.input_like = []
        self.output_like = [ttype]

    @staticmethod
    def n_input():
        return 0

    @staticmethod
    def n_output():
        return 1

    def __repr__(self):
        return f"Placeholder({self.ttype})"

    def to_const(self):
        const_node = Constant(self.ttype.ndims)
        const_node.abs_tensor = self.ttype
        return const_node

    def to_input(self):
        input_node = Input(self.ttype.ndims)
        input_node.abs_tensor = self.ttype
        return input_node

    def __str__(self):
        return "Placeholder"


# FIXME: Div will cause fuzzing crash. No integer to avoid division by zero.
@mark_materialize("core")
class Div(BcastBinaryOp):
    in_dtypes = [(i, i) for i in DTYPE_GEN_FLOATS]
    out_dtypes = [(i,) for i in DTYPE_GEN_FLOATS]


@mark_materialize("core", limit_domain=True)
class Pow(BcastBinaryOp):
    in_dtypes = [(i, i) for i in DTYPE_GEN_FLOATS]
    out_dtypes = [(i,) for i in DTYPE_GEN_FLOATS]


@mark_materialize("core")
class GELU(ElementWiseUnaryOp):
    in_dtypes = [(i,) for i in DTYPE_GEN_FLOATS]
    out_dtypes = [(i,) for i in DTYPE_GEN_FLOATS]


@mark_materialize("core")
class LeakyReLU(ElementWiseUnaryOp):
    in_dtypes = [(i,) for i in DTYPE_GEN_FLOATS]
    out_dtypes = [(i,) for i in DTYPE_GEN_FLOATS]

    def __init__(self):
        """See https://pytorch.org/docs/stable/generated/torch.nn.LeakyReLU.html"""
        super().__init__()


@mark_materialize("core")
class PReLU(ElementWiseUnaryOp):
    in_dtypes = [(DType.float32,)]
    out_dtypes = [(DType.float32,)]


@mark_materialize("core")
class Sigmoid(ElementWiseUnaryOp):
    in_dtypes = [(i,) for i in DTYPE_GEN_FLOATS]
    out_dtypes = [(i,) for i in DTYPE_GEN_FLOATS]


class TrigonometricOp(ElementWiseUnaryOp):
    pass


@mark_materialize("core")
class Sin(TrigonometricOp):
    in_dtypes = [(i,) for i in DTYPE_GEN_FLOATS]
    out_dtypes = [(i,) for i in DTYPE_GEN_FLOATS]


@mark_materialize("core")
class Cos(TrigonometricOp):
    in_dtypes = [(i,) for i in DTYPE_GEN_FLOATS]
    out_dtypes = [(i,) for i in DTYPE_GEN_FLOATS]


@mark_materialize("core", limit_domain=True)
class Asin(TrigonometricOp):
    in_dtypes = [(i,) for i in DTYPE_GEN_FLOATS]
    out_dtypes = [(i,) for i in DTYPE_GEN_FLOATS]


@mark_materialize("core", limit_domain=True)
class Acos(TrigonometricOp):
    in_dtypes = [(i,) for i in DTYPE_GEN_FLOATS]
    out_dtypes = [(i,) for i in DTYPE_GEN_FLOATS]


@mark_materialize("core")
class Tan(TrigonometricOp):
    in_dtypes = [(i,) for i in DTYPE_GEN_FLOATS]
    out_dtypes = [(i,) for i in DTYPE_GEN_FLOATS]


@mark_materialize("core")
class Atan(TrigonometricOp):
    in_dtypes = [(i,) for i in DTYPE_GEN_FLOATS]
    out_dtypes = [(i,) for i in DTYPE_GEN_FLOATS]


@mark_materialize("core")
class Abs(ElementWiseUnaryOp):
    in_dtypes = [(i,) for i in DTYPE_GEN_NON_BOOL if i != DType.uint8]
    in_dtypes = [
        (i,)
        for i in DTYPE_GEN_NON_BOOL
        if i not in {DType.complex64, DType.complex128, DType.uint8}
    ]

    def type_transfer(self, input_shapes: List[AbsTensor]) -> List[AbsTensor]:
        if input_shapes[0].dtype == DType.complex64:
            return [AbsTensor(input_shapes[0].shape, DType.float32)]
        elif input_shapes[0].dtype == DType.complex128:
            return [AbsTensor(input_shapes[0].shape, DType.float64)]
        return super().type_transfer(input_shapes)


@mark_materialize("core")
class ReLU(ElementWiseUnaryOp):
    in_dtypes = [(i,) for i in DTYPE_GEN_FLOATS]
    out_dtypes = [(i,) for i in DTYPE_GEN_FLOATS]


@mark_materialize("core")
class Ceil(ElementWiseUnaryOp):
    in_dtypes = [(i,) for i in DTYPE_GEN_FLOATS]
    out_dtypes = [(i,) for i in DTYPE_GEN_FLOATS]


@mark_materialize("core")
class Floor(ElementWiseUnaryOp):
    in_dtypes = [(i,) for i in DTYPE_GEN_FLOATS]
    out_dtypes = [(i,) for i in DTYPE_GEN_FLOATS]


@mark_materialize("core")
class Clip(ElementWiseUnaryOp):
    in_dtypes = [(i,) for i in DTYPE_GEN_NON_BOOL]


@mark_materialize("core")
class Round(ElementWiseUnaryOp):
    in_dtypes = [(i,) for i in DTYPE_GEN_FLOATS]
    out_dtypes = [(i,) for i in DTYPE_GEN_FLOATS]


@mark_materialize("core", limit_domain=True)
class Sqrt(ElementWiseUnaryOp):
    in_dtypes = [(i,) for i in DTYPE_GEN_FLOATS]
    out_dtypes = [(i,) for i in DTYPE_GEN_FLOATS]


@mark_materialize("core", limit_domain=True)
class Log2(ElementWiseUnaryOp):
    in_dtypes = [(i,) for i in DTYPE_GEN_FLOATS]
    out_dtypes = [(i,) for i in DTYPE_GEN_FLOATS]


@mark_materialize("core")
class Neg(ElementWiseUnaryOp):
    in_dtypes = [(i,) for i in DTYPE_GEN_NON_BOOL]
    out_dtypes = [(i,) for i in DTYPE_GEN_NON_BOOL]


@mark_materialize("core")
class Softmax(ElementWiseUnaryOp):
    in_dtypes = [(i,) for i in DTYPE_GEN_FLOATS]
    out_dtypes = [(i,) for i in DTYPE_GEN_FLOATS]

    def __init__(self, dim: Union[int, z3.ExprRef]):
        super().__init__()
        self.dim = dim
        self.inp_ranks = [rank_from(1)]
        self.out_ranks = [rank_from(1)]

    def requires(self, input_shapes: List[AbsTensor]) -> List[Union[z3.BoolRef, bool]]:
        return [nnsmith_lt(self.dim, input_shapes[0].ndims), nnsmith_ge(self.dim, 0)]


class Pool2d(UnaryOpBase):
    # TODO: distinguish stride_h and stride_w
    in_dtypes = [(i,) for i in DTYPE_GEN_FLOATS]
    out_dtypes = [(i,) for i in DTYPE_GEN_FLOATS]

    def __init__(
        self,
        kernel_h_size: Union[int, z3.ExprRef],
        kernel_w_size: Union[int, z3.ExprRef],
        stride: Union[int, z3.ExprRef],
        padding: Union[int, z3.ExprRef],
    ):
        super().__init__()
        self.kernel_h_size = kernel_h_size
        self.kernel_w_size = kernel_w_size
        self.stride = stride
        self.padding = padding

        self.inp_ranks = [(4,)]  # NCHW
        self.out_ranks = [(4,)]  # NCHW

    def type_transfer(self, input_shapes: List[AbsTensor]) -> List[AbsTensor]:

        abs_tensor = AbsTensor([], dtype=input_shapes[0].dtype)
        # Batch dim: just copy
        abs_tensor.shape.append(input_shapes[0].shape[0])
        # Output channels
        abs_tensor.shape.append(input_shapes[0].shape[1])
        abs_tensor.shape.append(
            (
                nnsmith_div(
                    nnsmith_add(
                        nnsmith_sub(input_shapes[0].shape[2], self.kernel_h_size),
                        2 * self.padding,
                    ),
                    self.stride,
                )
                + 1
            )
        )
        abs_tensor.shape.append(
            (
                nnsmith_div(
                    nnsmith_add(
                        nnsmith_sub(input_shapes[0].shape[3], self.kernel_w_size),
                        2 * self.padding,
                    ),
                    self.stride,
                )
                + 1
            )
        )
        return [abs_tensor]

    def requires(self, input_shapes):
        cons = []
        ret = []
        cons.append(nnsmith_ge(self.kernel_h_size, 1))
        cons.append(nnsmith_ge(self.kernel_w_size, 1))
        cons.append(
            nnsmith_le(
                self.kernel_h_size,
                nnsmith_add(input_shapes[0].shape[2], 2 * self.padding),
            )
        )
        cons.append(
            nnsmith_le(
                self.kernel_w_size,
                nnsmith_add(input_shapes[0].shape[3], 2 * self.padding),
            )
        )
        cons.append(nnsmith_ge(self.stride, 1))
        cons.append(nnsmith_ge(self.padding, 0))
        # not too extream to avoid torch exporter issue
        cons.append(nnsmith_le(self.padding, 255))
        cons.append(nnsmith_le(self.padding, nnsmith_div(self.kernel_h_size, 2)))
        cons.append(nnsmith_le(self.padding, nnsmith_div(self.kernel_w_size, 2)))
        cons.extend(
            [nnsmith_gt(v, 0) for v in input_shapes[0].shape]
        )  # dim cannot be 0 for maxpool

        for c in cons:
            ret.append(c)
        return ret

    def flops(self, input_shapes):
        return nnsmith_mul(
            nnsmith_mul(
                self.checked_type_transfer(input_shapes)[0].nelement(),
                self.kernel_h_size,
            ),
            self.kernel_w_size,
        )

    def deduct_inp_ranks_and_dtype(
        self, out_abs_tensor: List[AbsTensor]
    ) -> List[Tuple[int, DType]]:
        return [(4, out_abs_tensor[0].dtype)]


@mark_materialize("core")
class MaxPool2d(Pool2d):
    pass


@mark_materialize("core")
class AvgPool2d(Pool2d):
    pass


@mark_materialize("core")
class Slice(UnaryOpBase):
    # pytorch slice always exported as a stack of single-dim slices, so only model sinlge-dim slice here
    # pytorch slice only supports forward slicing, so only model forward slicing here
    in_dtypes = [(i,) for i in DTYPE_GEN_ALL]
    INT_MAX = 2**63 - 1
    INT_MIN = -(2**63)

    def __init__(self, start, end, step):
        super().__init__()
        self.inp_ranks = [rank_from(1)]
        self.out_ranks = [rank_from(1)]
        self.start = start
        self.end = end
        self.step = step

    def __str__(self) -> str:
        if "axis" in self.extra_attrs:
            tail = {
                "axis": self.extra_attrs["axis"],
                "region": self.extra_attrs["region"],
            }
        else:
            tail = {}
        if isinstance(self.start, int):
            tail["start"] = self.start
        if isinstance(self.end, int):
            tail["end"] = self.end
        if isinstance(self.step, int):
            tail["step"] = self.step
        return self.name() + " " + str(tail).replace(":", "=")

    def _get_attrs(self, ndims):
        ConstraintCheck.true(ndims > 0)
        if "axis" not in self.extra_attrs:
            self.extra_attrs["ndims"] = ndims
            self.extra_attrs["axis"] = random.randint(0, ndims - 1)
            # specifying the region of the start and end pointer.
            # start \in [0, dim_s-1] if region=='right' else [-dim_s, -1]
            # end \in [-dim_s, -1] if region=='left' else [0, dim_s]
            self.extra_attrs["region"] = random.choice(["left", "mid", "right"])
            if random.uniform(0, 1) < 0.1:
                # torch exporter does not support start=INT_MIN
                self.end = self.INT_MAX
        return self.extra_attrs["axis"]

    def requires(self, input_shapes: List[AbsTensor]):
        inp = input_shapes[0]
        axis = self._get_attrs(inp.ndims)
        reg = self.extra_attrs["region"]
        cons = []
        dim_s = inp.shape[axis]
        # range for start
        l, r = (0, nnsmith_sub(dim_s, 1))
        # range for end
        ll, rr = (0, dim_s)
        assert not isinstance(self.start, int)
        cons.append(
            z3.And(  # start \in [l, r]
                nnsmith_ge(self.start, l), nnsmith_le(self.start, r)
            )
        )
        if not isinstance(self.end, int):
            cons.append(
                z3.And(  # end \in [ll, rr]
                    nnsmith_ge(self.end, ll), nnsmith_le(self.end, rr)
                )
            )
            cons.append(nnsmith_gt(self.end, self.start))
        else:
            assert self.end == self.INT_MAX

        cons.append(nnsmith_ge(self.step, 1))  # forward slicing only
        cons.append(nnsmith_le(self.step, dim_s))
        return cons

    def type_transfer(self, input_shapes: List[AbsTensor]) -> List[AbsTensor]:
        inp = input_shapes[0]
        axis = self._get_attrs(inp.ndims)
        s = list(inp.shape)
        end = self.end
        if self.end == Slice.INT_MAX:
            end = inp.shape[axis]
        s[axis] = nnsmith_div(
            nnsmith_add(nnsmith_sub(end, self.start), nnsmith_sub(self.step, 1)),
            self.step,
        )
        return [AbsTensor(s, input_shapes[0].dtype)]

    def deduct_inp_ranks_and_dtype(
        self, out_abs_tensor: List[AbsTensor]
    ) -> List[Tuple[int, DType]]:
        return [(out_abs_tensor[0].ndims, out_abs_tensor[0].dtype)]


def _pad_num_var_param(rstart=1, max=None):
    r = rstart  # rank
    ret = []
    while r <= __MAX_RANK__:
        h = r * 2
        if max is not None and h > max:
            break
        ret.append(h)
        r += 1
    return ret


class Pad(UnaryOpBase):
    num_var_param = _pad_num_var_param()
    in_dtypes = [(i,) for i in DTYPE_GEN_FLOATS]
    out_dtypes = [(i,) for i in DTYPE_GEN_FLOATS]

    def __str__(self) -> str:
        return f"{self.name()} (padding={list(self.padding_list)})"

    def __init__(self, padding_list, pad_t):
        super().__init__()
        self.padding_list = padding_list
        self.extra_attrs["type"] = pad_t
        self.inp_ranks = [rank_from(len(padding_list) // 2)]
        self.out_ranks = [rank_from(len(padding_list) // 2)]
        assert (
            len(self.padding_list) % 2 == 0
        ), f"padding_list must be even, got {self.padding_list}"

    def requires(self, input_shapes: List[AbsTensor]) -> List[Union[z3.BoolRef, bool]]:
        pad = self.padding_list
        isv = input_shapes[0].shape
        cons = []
        for i in range(len(pad) // 2):
            j = len(isv) - 1 - i
            # When using negative padding, neither side should erase more than the original size
            cons.append(nnsmith_gt(nnsmith_add(pad[i * 2], isv[j]), 0))
            cons.append(nnsmith_gt(nnsmith_add(pad[i * 2 + 1], isv[j]), 0))
            cons.append(
                nnsmith_gt(
                    nnsmith_add(pad[i * 2 + 1], nnsmith_add(pad[i * 2], isv[j])), 0
                )
            )
        for s in input_shapes[0].shape[1:]:
            cons.append(nnsmith_gt(s, 0))
        return cons

    def type_transfer(self, input_shapes: List[AbsTensor]) -> List[AbsTensor]:
        isv = input_shapes[0].shape
        pad = self.padding_list
        s = list(isv)
        for i in range(len(pad) // 2):
            j = len(isv) - 1 - i
            s[j] = nnsmith_add(nnsmith_add(s[j], pad[i * 2]), pad[i * 2 + 1])
        return [AbsTensor(s, input_shapes[0].dtype)]

    def deduct_inp_ranks_and_dtype(
        self, out_abs_tensor: List[AbsTensor]
    ) -> List[Tuple[int, DType]]:
        return [(out_abs_tensor[0].ndims, out_abs_tensor[0].dtype)]


@mark_materialize("core")
class ConstPad(Pad):
    def __init__(self, *padding_list):
        super().__init__(padding_list, "constant")


@mark_materialize("core")
class ReplicatePad(Pad):
    num_var_param = _pad_num_var_param(2, max=6)

    def __init__(self, *padding_list):
        super().__init__(padding_list, "replicate")
        self.inp_ranks = [rank_range(len(padding_list) // 2 + 1, 4)]
        self.out_ranks = [rank_range(len(padding_list) // 2 + 1, 4)]


@mark_materialize("core")
class ReflectPad(Pad):
    num_var_param = _pad_num_var_param(2, max=6)
    in_dtypes = [(i,) for i in DTYPE_GEN_FLOATS if i != DType.float16]
    out_dtypes = [(i,) for i in DTYPE_GEN_FLOATS if i != DType.float16]

    def __init__(self, *padding_list):
        super().__init__(padding_list, "reflect")
        self.inp_ranks = [rank_range(len(padding_list) // 2 + 1, 4)]
        self.out_ranks = [rank_range(len(padding_list) // 2 + 1, 4)]

    def requires(self, input_shapes: List[AbsTensor]) -> List[Union[z3.BoolRef, bool]]:
        cons = super().requires(input_shapes)
        pad = self.padding_list
        isv = input_shapes[0].shape
        for i in range(len(pad) // 2):
            j = len(isv) - 1 - i
            # per torch's complaint: Padding size should be less than the corresponding input dimension
            cons.append(nnsmith_lt(pad[i * 2], isv[j]))
            cons.append(nnsmith_lt(pad[i * 2 + 1], isv[j]))
            # same sign to avoid ORT bugs
            cons.append(nnsmith_ge(pad[i * 2] * pad[i * 2 + 1], 0))
        return cons


class Expand(UnaryOpBase, ABC):
    in_dtypes = [(i,) for i in DTYPE_GEN_ALL]
    out_dtypes = [(i,) for i in DTYPE_GEN_ALL]
    # expand_dim cannot be symbolic. So just expand it.

    def __init__(self, expand_last_dim: int, expand_n: Union[int, z3.ExprRef]):
        """See https://pytorch.org/docs/stable/generated/torch.Tensor.expand.html"""
        super().__init__()
        self.inp_ranks = [rank_all()]
        SanityCheck.ge(expand_last_dim, 1)
        self.expand_last_dim = expand_last_dim
        self.expand_n = expand_n

    def type_transfer(self, input_shapes: List[AbsTensor]) -> List[AbsTensor]:
        if self.expand_last_dim <= len(input_shapes[0].shape):
            # NOTE: Werid, deepcopy is useless here.
            shape = AbsTensor(
                shape=[*input_shapes[0].shape], dtype=input_shapes[0].dtype
            )
            shape.shape[-self.expand_last_dim] = self.expand_n
            return [shape]
        else:  # expand it;
            # for example. we have:
            #       input shape [u, v]
            #       expand_last_dim <- 4
            #       return [expand_n, 1, u, v] where `1` is padded.
            dtype = input_shapes[0].dtype
            return [
                AbsTensor(
                    [
                        self.expand_n,
                        *(
                            [1]
                            * (self.expand_last_dim - len(input_shapes[0].shape) - 1)
                        ),
                        *input_shapes[0].shape,
                    ],
                    dtype,
                )
            ]

    def requires(self, input_shapes):
        SanityCheck.ge(self.expand_last_dim, 1)

        input_shape = input_shapes[0].shape
        if self.expand_last_dim <= len(input_shape):  # index valid
            cons = [
                nnsmith_eq(input_shape[-self.expand_last_dim], 1),
                nnsmith_ge(self.expand_n, 1),
            ]
            return cons
        return [nnsmith_ge(self.expand_n, 1)]

    def deduct_inp_ranks_and_dtype(
        self, out_abs_tensor: List[AbsTensor]
    ) -> List[Tuple[int, DType]]:
        inp_rank = (
            self.expand_last_dim
            if out_abs_tensor[0].ndims < self.expand_last_dim
            else out_abs_tensor[0].ndims
        )
        ConstraintCheck.ge(out_abs_tensor[0].ndims, self.expand_last_dim)
        return [(inp_rank, out_abs_tensor[0].dtype)]


@mark_materialize("core")
class ExpandLast1(Expand):
    def __init__(self, expand_n: Union[int, z3.ExprRef]):
        super().__init__(expand_last_dim=1, expand_n=expand_n)


@mark_materialize("core")
class ExpandLast2(Expand):
    def __init__(self, expand_n: Union[int, z3.ExprRef]):
        super().__init__(expand_last_dim=2, expand_n=expand_n)


@mark_materialize("core")
class ExpandLast3(Expand):
    def __init__(self, expand_n: Union[int, z3.ExprRef]):
        super().__init__(expand_last_dim=3, expand_n=expand_n)


@mark_materialize("core")
class ExpandLast4(Expand):
    def __init__(self, expand_n: Union[int, z3.ExprRef]):
        super().__init__(expand_last_dim=4, expand_n=expand_n)


@mark_materialize("core")
class BatchNorm2d(ElementWiseUnaryOp):
    in_dtypes = [(DType.float32,)]
    out_dtypes = [(DType.float32,)]

    def __init__(self, nfeat):
        super().__init__()
        self.inp_ranks = [(4,)]
        self.out_ranks = [(4,)]
        self.nfeat = nfeat

    def deduct_inp_ranks_and_dtype(
        self, out_abs_tensor: List[AbsTensor]
    ) -> List[Tuple[int, DType]]:
        return [(4, DType.float32)]

    def requires(self, input_shapes: List[AbsTensor]) -> List[Union[z3.BoolRef, bool]]:
        return [
            nnsmith_eq(self.nfeat, input_shapes[0].shape[1]),
            nnsmith_ge(input_shapes[0].shape[0], 2),
        ]  # batch size = 1 -> fail training.


@mark_materialize("core")
class Conv1d(UnaryOpBase):
    in_dtypes = [(DType.float32,)]
    out_dtypes = [(DType.float32,)]

    def __init__(
        self,
        in_channels: Union[int, z3.ExprRef],
        out_channels: Union[int, z3.ExprRef],
        kernel_size: Union[int, z3.ExprRef],
        stride: Union[int, z3.ExprRef],
        padding: Union[int, z3.ExprRef],
        dilation: Union[int, z3.ExprRef],
    ):
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = kernel_size
        self.stride = stride
        self.padding = padding
        self.dilation = dilation

        self.inp_ranks = [(3,)]  # NCL
        self.out_ranks = [(3,)]  # NCL

    def type_transfer(self, input_shapes: List[AbsTensor]) -> List[AbsTensor]:
        abs_tensor = AbsTensor(
            [input_shapes[0].shape[0], self.out_channels], dtype=input_shapes[0].dtype
        )
        mimic_k = self.kernel_size + (self.dilation - 1) * (self.kernel_size - 1)
        abs_tensor.shape.append(
            (
                nnsmith_div(
                    nnsmith_add(
                        nnsmith_sub(input_shapes[0].shape[2], mimic_k), 2 * self.padding
                    ),
                    self.stride,
                )
                + 1
            )
        )

        return [abs_tensor]

    def requires(self, input_shapes):
        # FIXME: Handling flops.
        cons = []
        cons.append(nnsmith_eq(self.in_channels, input_shapes[0].shape[1]))
        cons.append(nnsmith_ge(self.out_channels, 1))
        cons.append(nnsmith_ge(self.dilation, 1))
        mimic_k = self.kernel_size + (self.dilation - 1) * (self.kernel_size - 1)
        cons.append(nnsmith_ge(mimic_k, 1))
        cons.append(nnsmith_ge(self.stride, 1))
        cons.append(nnsmith_ge(self.padding, 0))
        cons.append(
            nnsmith_le(mimic_k, nnsmith_add(input_shapes[0].shape[2], 2 * self.padding))
        )
        # not too extream to avoid torch exporter issue
        cons.append(nnsmith_le(self.padding, 255))
        return cons

    def deduct_inp_ranks_and_dtype(
        self, out_abs_tensor: List[AbsTensor]
    ) -> List[Tuple[int, DType]]:
        return [(3, out_abs_tensor[0].dtype)]

    def __repr__(self) -> str:
        repr = f"Conv1d({self.in_channels}, {self.out_channels}, k={self.kernel_size}"
        if not isinstance(self.stride, int) or self.stride != 1:
            repr += f", s={self.stride}"
        if not isinstance(self.padding, int) or self.padding != 0:
            repr += f", p={self.padding}"
        if not isinstance(self.dilation, int) or self.dilation != 1:
            repr += f", d={self.dilation}"
        repr += ")"
        return repr


@mark_materialize("core")
class NCHWConv2d(UnaryOpBase):
    # FIXME: torch exporter does not support float64, may miss bugs
    in_dtypes = [(DType.float32,)]
    out_dtypes = [(DType.float32,)]

    def __init__(
        self,
        in_channels: Union[int, z3.ExprRef],
        out_channels: Union[int, z3.ExprRef],
        kernel_h_size: Union[int, z3.ExprRef],
        kernel_w_size: Union[int, z3.ExprRef],
        stride: Union[int, z3.ExprRef],
        padding: Union[int, z3.ExprRef],
        dilation_h: Union[int, z3.ExprRef],
        dilation_w: Union[int, z3.ExprRef],
    ):
        """See https://pytorch.org/docs/stable/generated/torch.nn.Conv2d.html"""
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_h_size = kernel_h_size
        self.kernel_w_size = kernel_w_size
        self.stride = stride
        self.padding = padding
        self.dilation_h = dilation_h
        self.dilation_w = dilation_w

        self.inp_ranks = [(4,)]  # NC(H,)W
        self.out_ranks = [(4,)]  # NC(H,)W

    def type_transfer(self, input_shapes: List[AbsTensor]) -> List[AbsTensor]:
        abs_tensor = AbsTensor(
            [input_shapes[0].shape[0], self.out_channels], dtype=input_shapes[0].dtype
        )

        mimic_kh = self.kernel_h_size + (self.dilation_h - 1) * (self.kernel_h_size - 1)
        mimic_kw = self.kernel_w_size + (self.dilation_w - 1) * (self.kernel_w_size - 1)

        abs_tensor.shape.append(
            (
                nnsmith_div(
                    nnsmith_add(
                        nnsmith_sub(input_shapes[0].shape[2], mimic_kh),
                        2 * self.padding,
                    ),
                    self.stride,
                )
                + 1
            )
        )
        abs_tensor.shape.append(
            (
                nnsmith_div(
                    nnsmith_add(
                        nnsmith_sub(input_shapes[0].shape[3], mimic_kw),
                        2 * self.padding,
                    ),
                    self.stride,
                )
                + 1
            )
        )
        return [abs_tensor]

    def requires(self, input_shapes):
        cons = []
        # TODO: Use eager mode for debugging.
        cons.append(nnsmith_eq(self.in_channels, input_shapes[0].shape[1]))
        cons.append(nnsmith_ge(self.out_channels, 1))
        cons.append(nnsmith_ge(self.dilation_h, 1))
        cons.append(nnsmith_ge(self.dilation_w, 1))
        mimic_kh = self.kernel_h_size + (self.dilation_h - 1) * (self.kernel_h_size - 1)
        mimic_kw = self.kernel_w_size + (self.dilation_w - 1) * (self.kernel_w_size - 1)
        cons.append(nnsmith_ge(mimic_kh, 1))
        cons.append(nnsmith_ge(mimic_kw, 1))
        cons.append(nnsmith_ge(self.stride, 1))
        cons.append(nnsmith_ge(self.padding, 0))
        cons.append(
            nnsmith_le(
                mimic_kh, nnsmith_add(input_shapes[0].shape[2], 2 * self.padding)
            )
        )
        cons.append(
            nnsmith_le(
                mimic_kw, nnsmith_add(input_shapes[0].shape[3], 2 * self.padding)
            )
        )
        # not too extream to avoid torch exporter issue
        cons.append(nnsmith_le(self.padding, 255))
        return cons

    def flops(self, input_shapes):
        w = AbsTensor(
            [
                self.out_channels,
                self.in_channels,
                self.kernel_h_size,
                self.kernel_w_size,
            ],
            dtype=input_shapes[0].dtype,
        )
        return nnsmith_mul(
            nnsmith_mul(
                nnsmith_mul(
                    self.type_transfer(input_shapes)[0].nelement(), self.in_channels
                ),
                self.kernel_h_size,
            ),
            self.kernel_w_size,
        )

    def n_floats(self, input_shapes):
        # FIXME: maybe need to take dilation into account?
        padded_data = AbsTensor(input_shapes[0].shape, dtype=input_shapes[0].dtype)
        padded_data.shape[2] = nnsmith_add(
            padded_data.shape[2], nnsmith_mul(2, self.padding)
        )
        padded_data.shape[3] = nnsmith_add(
            padded_data.shape[3], nnsmith_mul(2, self.padding)
        )
        w = AbsTensor(
            [
                self.out_channels,
                self.in_channels,
                self.kernel_h_size,
                self.kernel_w_size,
            ],
            dtype=input_shapes[0].dtype,
        )
        outs = super().n_floats(input_shapes)
        return nnsmith_add(nnsmith_add(w.nelement(), padded_data.nelement()), outs)

    def deduct_inp_ranks_and_dtype(
        self, out_abs_tensor: List[AbsTensor]
    ) -> List[Tuple[int, DType]]:
        return [(4, out_abs_tensor[0].dtype)]

    def __repr__(self) -> str:
        repr = f"Conv2d({self.in_channels}, {self.out_channels}, k=({self.kernel_h_size},{self.kernel_w_size})"
        if not isinstance(self.stride, int) or self.stride != 1:
            repr += f", s={self.stride}"
        if not isinstance(self.padding, int) or self.padding != 0:
            repr += f", p={self.padding}"
        if (
            not isinstance(self.dilation_h, int)
            or self.dilation_h != 1
            or self.dilation_w != 1
        ):
            repr += f", d=({self.dilation_h}, {self.dilation_w})"
        repr += ")"
        return repr


def random_group(n, k):
    xs = sorted([random.randint(0, n - k) for _ in range(k - 1)])
    xs = [0] + xs + [n - k]
    ret = []
    perm = list(range(n))
    random.shuffle(perm)
    for i in range(k):
        st = xs[i] + i
        ed = xs[i + 1] + i + 1
        assert st < ed, (xs, st, ed)
        assert ed <= n, (st, ed, n)
        assert st >= 0, (st, ed, n)
        ret.append([perm[j] for j in range(st, ed)])
    return ret


@mark_materialize("core")
class Reshape(UnaryOpBase):
    num_var_param = rank_range(1, 4)
    in_dtypes = [(i,) for i in DTYPE_GEN_ALL]
    out_dtypes = [(i,) for i in DTYPE_GEN_ALL]

    def __init__(self, *target_shape):
        super().__init__()
        self.inp_ranks = [rank_range(1, 4)]
        self.out_ranks = [(len(target_shape),)]
        self.target_shape: List[Union[int, z3.ExprRef]] = list(target_shape)

    def type_transfer(self, input_shapes: List[AbsTensor]) -> List[AbsTensor]:
        __MAX_SOLVE_SYMBOL__ = 8
        # otherwise OOM.
        ConstraintCheck.le(
            input_shapes[0].ndims + len(self.target_shape), __MAX_SOLVE_SYMBOL__
        )

        if -1 not in self.target_shape:
            return [AbsTensor(self.target_shape, dtype=input_shapes[0].dtype)]
        # else
        abs_tensor = AbsTensor(self.target_shape, dtype=input_shapes[0].dtype)
        auto_dim = -1
        accum = 1
        for i, v in enumerate(self.target_shape):
            if v == -1:
                SanityCheck.eq(auto_dim, -1)
                auto_dim = i
            else:
                accum = nnsmith_mul(accum, v)

        abs_tensor.shape[auto_dim] = nnsmith_div(
            reduce(lambda x, y: nnsmith_mul(x, y), input_shapes[0].shape, 1), accum
        )

        return [abs_tensor]

    def requires(self, input_shapes):
        ret = []

        inp = input_shapes[0]
        src_len, dst_len = inp.ndims, len(self.target_shape)
        if src_len == 0:
            src_len = 1  # special handling for scalar
        if dst_len == 0:
            dst_len = 1  # special handling for scalar
        gres_config = os.getenv("NNSMITH_GRES", "4")
        if gres_config == "5":
            ng = 1
        elif gres_config == "3":
            ng = min(src_len, dst_len)
        elif gres_config == "4":
            ub = min(src_len, dst_len)
            ng = random.choices(
                range(1, ub + 1), k=1, weights=[2**i for i in range(ub)]
            )[0]
        else:
            raise ValueError(f"NNSMITH_GRES={gres_config} is not recognized")
        src_group = random_group(src_len, ng)
        dst_group = random_group(dst_len, ng)
        self.ng = ng
        self.src_group = src_group
        self.dst_group = dst_group
        assert len(src_group) == len(dst_group) == ng, (src_group, dst_group)

        # group constraints
        src_vars = inp.shape
        dst_vars = self.target_shape
        if len(src_vars) == 0:
            src_vars = [1]  # special handling for scalar
        if len(dst_vars) == 0:
            dst_vars = [1]  # special handling for scalar
        cons_group = []
        for gid in range(ng):
            src_idx = src_group[gid]
            dst_idx = dst_group[gid]
            src_prod = reduce(nnsmith_mul, [src_vars[i] for i in src_idx], 1)
            dst_prod = reduce(nnsmith_mul, [dst_vars[i] for i in dst_idx], 1)
            cons_group.append(nnsmith_eq(src_prod, dst_prod))

        ret.extend(cons_group)
        if os.getenv("NNSMITH_CONS_RESHAPE", "off") != "off":
            # should not be too extreme!
            __DIM_LIMIT__ = 4096
            lim = __DIM_LIMIT__
            for s in self.target_shape[::-1]:
                ret.append(nnsmith_le(s, lim))
                lim //= 2
                lim = max(lim, 1)
        assert -1 not in self.target_shape
        return ret

    def deduct_inp_ranks_and_dtype(
        self, out_abs_tensor: List[AbsTensor]
    ) -> List[Tuple[int, DType]]:
        return [(-1, out_abs_tensor[0].dtype)]


@mark_materialize("core")
class Transpose(UnaryOpBase):
    in_dtypes = [(i,) for i in DTYPE_GEN_ALL]

    def __init__(self):
        super().__init__()
        self.inp_ranks = [rank_from(2)]
        self.out_ranks = [rank_from(2)]

    def _init_swap_dims(self, input_shape: List[Union[int, z3.ExprRef]]):
        ConstraintCheck.ge(len(input_shape), 2)
        self.inp_ranks = [len(input_shape)]
        if "dim0" not in self.extra_attrs or "dim1" not in self.extra_attrs:
            max_dim = len(input_shape) - 1
            self.extra_attrs["dim0"] = random.randint(0, max_dim)
            self.extra_attrs["dim1"] = (
                random.randint(1, max_dim) + self.extra_attrs["dim0"]
            ) % (1 + max_dim)
        return self.extra_attrs["dim0"], self.extra_attrs["dim1"]

    def type_transfer(self, input_shapes: List[AbsTensor]) -> List[AbsTensor]:
        dim0, dim1 = self._init_swap_dims(input_shapes[0].shape)
        shape = list(input_shapes[0].shape)
        shape[dim0], shape[dim1] = shape[dim1], shape[dim0]
        return [AbsTensor(shape, input_shapes[0].dtype)]

    def requires(self, input_shapes):
        dim0, dim1 = self._init_swap_dims(input_shapes[0].shape)
        SanityCheck.ge(
            len(input_shapes[0].shape),
            max(dim0, dim1) + 1,
            f"dim={len(input_shapes[0].shape)}.transpose({dim0},{dim1})",
        )
        return []

    def deduct_inp_ranks_and_dtype(
        self, out_abs_tensor: List[AbsTensor]
    ) -> List[Tuple[int, DType]]:
        return [(out_abs_tensor[0].ndims, out_abs_tensor[0].dtype)]


class InterpBase(UnaryOpBase):
    num_var_param = rank_range(1, 3)

    in_dtypes = [(i,) for i in DTYPE_GEN_FLOATS]
    out_dtypes = [(i,) for i in DTYPE_GEN_FLOATS]

    def __init__(self, *size):
        super().__init__()
        self.size = list(size)
        self.inp_ranks = [(len(size) + 2,)]
        self.out_ranks = [(len(size) + 2,)]

    def requires(self, input_shapes: List[AbsTensor]):
        return [nnsmith_gt(v, 0) for v in self.size] + [
            nnsmith_gt(v, 0) for v in input_shapes[0].shape
        ]

    def type_transfer(self, input_shapes: List[AbsTensor]) -> List[AbsTensor]:
        shape = list(input_shapes[0].shape)
        for i in range(len(self.size)):
            shape[-(1 + i)] = self.size[-(1 + i)]
        return [AbsTensor(shape, input_shapes[0].dtype)]

    def deduct_inp_ranks_and_dtype(
        self, out_abs_tensor: List[AbsTensor]
    ) -> List[Tuple[int, DType]]:
        return [(out_abs_tensor[0].ndims, out_abs_tensor[0].dtype)]


@mark_materialize("core")
class NearestInterp(InterpBase):
    pass


@mark_materialize("core")
class LinearInterp(InterpBase):
    num_var_param = [1]


@mark_materialize("core")
class BilinearInterp(InterpBase):
    num_var_param = [2]


@mark_materialize("core")
class BicubicInterp(InterpBase):
    num_var_param = [2]


@mark_materialize("core")
class TrilinearInterp(InterpBase):
    num_var_param = [3]


class ReduceBase(UnaryOpBase, ABC):
    _reduce_out_dtype = None  # None means same as input dtype

    def __init__(self):
        super().__init__()
        self.inp_ranks = [rank_from(1)]  # TVM bug ~ crash on scalar.min()
        self.out_ranks = [rank_range(0, __MAX_RANK__ - 1)]

    def __str__(self) -> str:
        return (
            self.name()
            + f'(dim={self.extra_attrs["reduce_dim"] if "reduce_dim" in self.extra_attrs else None})'
        )

    def _init_reduce_dim(self, input_shape: List[Union[int, z3.ExprRef]]):
        if "reduce_dim" not in self.extra_attrs:
            if len(input_shape) == 0:
                self.extra_attrs["reduce_dim"] = None
            else:
                self.extra_attrs["reduce_dim"] = random.randint(0, len(input_shape) - 1)
        return self.extra_attrs["reduce_dim"]

    def type_transfer(self, input_shapes: List[AbsTensor]) -> List[AbsTensor]:
        svar_list = []
        dim_idx = self._init_reduce_dim(input_shapes[0].shape)
        SanityCheck.lt(dim_idx, input_shapes[0].ndims)
        for i, v in enumerate(input_shapes[0].shape):
            if i != dim_idx:
                svar_list.append(v)
        return [
            AbsTensor(
                svar_list,
                input_shapes[0].dtype
                if self._reduce_out_dtype is None
                else self._reduce_out_dtype,
            )
        ]

    def requires(self, input_shapes: List[AbsTensor]):
        dim_idx = self._init_reduce_dim(input_shapes[0].shape)
        return [nnsmith_neq(input_shapes[0].shape[dim_idx], 0)]

    def _get_irank(self, orank):
        return orank + 1

    def deduct_inp_ranks_and_dtype(
        self, out_abs_tensor: List[AbsTensor]
    ) -> List[Tuple[int, DType]]:
        return [(self._get_irank(out_abs_tensor[0].ndims), out_abs_tensor[0].dtype)]


@mark_materialize("core")
class Squeeze(ReduceBase):
    in_dtypes = [(i,) for i in DTYPE_GEN_ALL]

    def requires(self, input_shapes):
        reduce_dim = self._init_reduce_dim(input_shapes[0].shape)
        if reduce_dim is None:
            return []
        return [nnsmith_eq(input_shapes[0].shape[reduce_dim], 1)]


@mark_materialize("core")
class ReduceSum(ReduceBase):
    in_dtypes = [(i,) for i in DTYPE_GEN_NON_BOOL]
    out_dtypes = [(i,) for i in DTYPE_GEN_NON_BOOL]


@mark_materialize("core")
class ReduceMin(ReduceBase):
    in_dtypes = [(i,) for i in DTYPE_GEN_NON_BOOL]
    out_dtypes = [(i,) for i in DTYPE_GEN_NON_BOOL]


@mark_materialize("core")
class ReduceMax(ReduceBase):
    in_dtypes = [(i,) for i in DTYPE_GEN_NON_BOOL]
    out_dtypes = [(i,) for i in DTYPE_GEN_NON_BOOL]


@mark_materialize("core")
class ReduceMean(ReduceBase):
    in_dtypes = [(i,) for i in DTYPE_GEN_FLOATS]
    out_dtypes = [(i,) for i in DTYPE_GEN_FLOATS]


@mark_materialize("core")
class ReduceProd(ReduceBase):
    in_dtypes = [(i,) for i in DTYPE_GEN_NON_BOOL]
    out_dtypes = [(i,) for i in DTYPE_GEN_NON_BOOL]


@mark_materialize("core")
class ArgMin(ReduceBase):
    in_dtypes = [(i,) for i in DTYPE_GEN_NON_BOOL if i not in DTYPE_GEN_COMPLEX]
    out_dtypes = [(DType.int64,)]
    _reduce_out_dtype = DType.int64

    def deduct_inp_ranks_and_dtype(
        self, out_abs_tensor: List[AbsTensor]
    ) -> List[Tuple[int, DType]]:
        return [
            (self._get_irank(out_abs_tensor[0].ndims), random.choice(self.in_dtypes)[0])
        ]


@mark_materialize("core")
class ArgMax(ReduceBase):
    in_dtypes = [(i,) for i in DTYPE_GEN_NON_BOOL if i not in DTYPE_GEN_COMPLEX]
    out_dtypes = [(DType.int64,)]
    _reduce_out_dtype = DType.int64

    def deduct_inp_ranks_and_dtype(
        self, out_abs_tensor: List[AbsTensor]
    ) -> List[Tuple[int, DType]]:
        return [
            (self._get_irank(out_abs_tensor[0].ndims), random.choice(self.in_dtypes)[0])
        ]


class TriBase(UnaryOpBase):
    in_dtypes = [(i,) for i in DTYPE_GEN_ALL]
    out_dtypes = [(i,) for i in DTYPE_GEN_ALL]

    def __init__(self, diagonal: Union[int, z3.ExprRef]):
        super().__init__()
        self.diagonal = diagonal
        # tril is only for 2-D matrix
        self.inp_ranks = [(2,)]
        self.out_ranks = [(2,)]

    def type_transfer(self, input_shapes: List[AbsTensor]) -> List[AbsTensor]:
        SanityCheck.eq(len(input_shapes), 1)
        return [input_shapes[0]]

    def deduct_inp_ranks_and_dtype(
        self, out_abs_tensor: List[AbsTensor]
    ) -> List[Tuple[int, DType]]:
        return [(2, out_abs_tensor[0].dtype)]


@mark_materialize("core")
class Tril(TriBase):
    def requires(self, input_shapes: List[AbsTensor]) -> List[Union[z3.BoolRef, bool]]:
        ConstraintCheck.true(input_shapes[0].ndims == 2)
        nrow = input_shapes[0].shape[0]
        ncol = input_shapes[0].shape[1]
        return [z3.And(self.diagonal >= -nrow, (ncol - 1) >= self.diagonal)]


@mark_materialize("core")
class Triu(TriBase):
    def requires(self, input_shapes: List[AbsTensor]) -> List[Union[z3.BoolRef, bool]]:
        ConstraintCheck.true(input_shapes[0].ndims == 2)
        nrow = input_shapes[0].shape[0]
        ncol = input_shapes[0].shape[1]
        return [z3.And(self.diagonal >= -(nrow - 1), ncol >= self.diagonal)]


class Concat(AbsOpBase):
    MAX_ARITY = 5
    MAX_RANK = 5
    out_dtypes = [(i,) for i in DTYPE_GEN_ALL]
    same_inp_dims = True

    def __str__(self) -> str:
        return "Concat " + str(self.extra_attrs).replace(":", "=")

    def __init__(self, arity):
        super().__init__()
        SanityCheck.le(arity, Concat.MAX_ARITY)
        self.arity = arity
        self.inp_ranks = [(rank_from(1))] * arity
        self.out_ranks = [(rank_from(1))]

    def _init_concat_axis(self, input_shapes: List[AbsTensor]) -> int:
        if "axis" not in self.extra_attrs:
            self.extra_attrs["axis"] = random.randint(0, input_shapes[0].ndims - 1)
        return self.extra_attrs["axis"]

    def requires(self, input_shapes: List[AbsTensor]) -> List[Union[z3.BoolRef, bool]]:
        ndims = input_shapes[0].ndims
        SanityCheck.gt(ndims, self._init_concat_axis(input_shapes))
        for s in input_shapes:
            SanityCheck.eq(s.ndims, ndims)
        cons = []
        for d in range(ndims):
            if d != self._init_concat_axis(input_shapes):
                cons.extend(
                    nnsmith_eq(s.shape[d], input_shapes[0].shape[d])
                    for s in input_shapes
                )
        return cons

    def type_transfer(self, input_shapes: List[AbsTensor]) -> List[AbsTensor]:
        SanityCheck.true(input_shapes[0].ndims > 0)
        axis = self._init_concat_axis(input_shapes)
        os = AbsTensor(input_shapes[0].shape, input_shapes[0].dtype)
        os.shape[axis] = reduce(nnsmith_add, [s.shape[axis] for s in input_shapes])
        return [os]

    def deduct_inp_ranks_and_dtype(
        self, out_abs_tensor: List[AbsTensor]
    ) -> List[Tuple[int, DType]]:
        return [
            (out_abs_tensor[0].ndims, out_abs_tensor[0].dtype)
            for _ in range(self.arity)
        ]


# the semantic of `in_dtypes` is not possible dtypes in "max rank". but simply in "rank". don't mess up the definition.
@mark_materialize("core")
class Concat1(Concat):
    in_dtypes = [(i,) for i in DTYPE_GEN_ALL]

    def __init__(self):
        super().__init__(1)


@mark_materialize("core")
class Concat2(Concat):
    in_dtypes = [(i, i) for i in DTYPE_GEN_ALL]

    def __init__(self):
        super().__init__(2)


@mark_materialize("core")
class Concat3(Concat):
    in_dtypes = [(i, i, i) for i in DTYPE_GEN_ALL]

    def __init__(self):
        super().__init__(3)


@mark_materialize("core")
class Concat4(Concat):
    in_dtypes = [(i, i, i, i) for i in DTYPE_GEN_ALL]

    def __init__(self):
        super().__init__(4)


@mark_materialize("core")
class Concat5(Concat):
    in_dtypes = [(i, i, i, i, i) for i in DTYPE_GEN_ALL]

    def __init__(self):
        super().__init__(5)


class Cast(ElementWiseUnaryOp, ABC):
    in_dtypes = [(i,) for i in DTYPE_GEN_ALL]

    def __init__(self, dtype):
        super().__init__()
        self.inp_ranks = [rank_all()]
        self.out_ranks = [rank_all()]
        self.extra_attrs = {"to": dtype}

    def __str__(self) -> str:
        return "Cast " + str(self.extra_attrs).replace(":", "=")

    def requires(self, input_shapes: List[AbsTensor]) -> List[Union[z3.BoolRef, bool]]:
        return []

    def type_transfer(self, input_shapes: List[AbsTensor]) -> List[AbsTensor]:
        assert len(input_shapes) == 1
        return [AbsTensor(input_shapes[0].shape, self.extra_attrs["to"])]

    def deduct_inp_ranks_and_dtype(
        self, out_abs_tensor: List[AbsTensor]
    ) -> List[Tuple[int, DType]]:
        return [(out_abs_tensor[0].ndims, self.extra_attrs["to"])]


@mark_materialize("core")
class CastF32(Cast):
    out_dtypes = [(DType.float32,)]

    def __init__(self):
        super().__init__(DType.float32)


@mark_materialize("core")
class CastF64(Cast):
    out_dtypes = [(DType.float64,)]

    def __init__(self):
        super().__init__(DType.float64)


@mark_materialize("core")
class CastI32(Cast):
    out_dtypes = [(DType.int32,)]

    def __init__(self):
        super().__init__(DType.int32)


@mark_materialize("core")
class CastI64(Cast):
    out_dtypes = [(DType.int64,)]

    def __init__(self):
        super().__init__(DType.int64)


@mark_materialize("core")
class CastBool(Cast):
    out_dtypes = [(DType.bool,)]

    def __init__(self):
        super().__init__(DType.bool)


class MatMul(BinaryOpBase):
    in_dtypes = [(i, i) for i in DTYPE_GEN_NON_BOOL if i not in DTYPE_GEN_COMPLEX]
    out_dtypes = [(i,) for i in DTYPE_GEN_NON_BOOL if i not in DTYPE_GEN_COMPLEX]

    def __init__(self):
        super().__init__()
        # Consider at most 3D tensors (batched mm)
        self.inp_ranks = [rank_range(1, 3), rank_range(1, 3)]
        self.out_ranks = [rank_until(3)]

    def type_transfer(self, input_shapes: List[AbsTensor]) -> List[AbsTensor]:
        # https://pytorch.org/docs/stable/generated/torch.matmul.html#torch.matmul
        # shape: [*batches(?), *rc (row and col)]
        lhs = input_shapes[0].shape
        rhs = input_shapes[1].shape

        lrc = lhs[-2:]
        rrc = rhs[-2:]
        orc = [*lrc[:-1], *rrc[1:]]

        lbatch = lhs[: -len(lrc)]
        rbatch = rhs[: -len(rrc)]
        batches = []
        if len(lbatch) > len(rbatch):
            batches = lbatch[: len(lbatch) - len(rbatch)]
            for x, y in zip(lbatch[len(batches) :], rbatch):
                batches.append(nnsmith_max(x, y))
        else:
            batches = rbatch[: len(rbatch) - len(lbatch)]
            for x, y in zip(lbatch, rbatch[len(batches) :]):
                batches.append(nnsmith_max(x, y))

        return [AbsTensor([*batches, *orc], input_shapes[0].dtype)]

    def requires(self, input_shapes: List[AbsTensor]) -> List[Union[z3.BoolRef, bool]]:
        cons = []

        lhs = input_shapes[0].shape
        rhs = input_shapes[1].shape

        lrc = lhs[-2:]
        rrc = rhs[-2:]

        # CHECK: l.cols = r.rows
        cons.append(lrc[-1] == rrc[0])

        # CHECK: batch dim broadcastable
        lbatch = lhs[: -len(lrc)]
        rbatch = rhs[: -len(rrc)]
        common_tail = min(len(lbatch), len(rbatch))
        for x, y in zip(lbatch[-common_tail:], rbatch[-common_tail:]):
            cons.append(nnsmith_or(x == y, nnsmith_or(x == 1, y == 1)))

        return cons

    def deduct_inp_ranks_and_dtype(
        self, out_abs_tensor: List[AbsTensor]
    ) -> List[Tuple[int, DType]]:
        # rank(a) = batch_rank(a) + rc_rank(a)
        # rank(b) = batch_rank(b) + rc_rank(b)
        # out_rank = max(br_a, br_b) + (rcr_a + rcr_b) - 2
        # 1 <= rcr_a or rcr_b <= min(2, out_rank + 2)

        # br_a = ranks[0], rcr_a = ranks[1]
        # br_b = ranks[2], rcr_b = ranks[3]
        ranks = [0, 1, 0, 1]

        def check_sat():
            return (
                out_abs_tensor[0].ndims
                == max(ranks[0], ranks[2]) + (ranks[1] + ranks[3]) - 2
            )

        while not check_sat():
            inc_candidates = []
            inc_candidates.append(1 if ranks[1] < 2 else 0)
            inc_candidates.append(3 if ranks[3] < 2 else 2)
            choice = random.choice(inc_candidates)
            ranks[choice] += 1

        return [
            (ranks[0] + ranks[1], out_abs_tensor[0].dtype),
            (ranks[2] + ranks[3], out_abs_tensor[0].dtype),
        ]


def concretize_op(
    op: Union[AbsOpBase, Placeholder], model: Optional[z3.ModelRef]
) -> Union[AbsOpBase, Placeholder]:
    if isinstance(op, Constant) or isinstance(op, Input):
        ret_op = deepcopy(op)
        values = []

        for idx, s in enumerate(op.abs_tensor.shape):
            if isinstance(s, z3.ExprRef):
                ret_op.abs_tensor.shape[idx] = model.eval(s).as_long()

        return ret_op
    elif isinstance(op, Placeholder):
        shape = []
        for idx, s in enumerate(op.ttype.shape):
            if isinstance(s, z3.ExprRef):
                shape.append(model.eval(s).as_long())
        return Placeholder(AbsTensor(shape=shape, dtype=op.ttype.dtype))

    # Non-inp / const types.
    construct_param_dict = signature(op.__init__).parameters
    values = []
    symbolic_idx = []

    if op.num_var_param is not None:
        # input is a variable list.
        key = list(construct_param_dict.keys())[0]
        values = list(getattr(op, key))
        symbolic_idx = [
            i for i in range(len(values)) if isinstance(values[i], z3.ExprRef)
        ]
    else:
        for idx, key in enumerate(construct_param_dict):
            param = getattr(op, key)
            values.append(param)
            if isinstance(param, z3.ExprRef):
                symbolic_idx.append(idx)

    for idx in symbolic_idx:
        values[idx] = model.eval(values[idx]).as_long()

    concrete_op = type(op)(*values)
    concrete_op.inp_ranks = op.inp_ranks
    concrete_op.out_ranks = op.out_ranks
    concrete_op.extra_attrs = op.extra_attrs

    return concrete_op


_PRAGMA_ONCE_CORE_OP = False



import os
import random
import time
import traceback
from importlib.util import module_from_spec, spec_from_file_location
from pathlib import Path
from types import FunctionType
from typing import Type

import hydra
from omegaconf import DictConfig

from nnsmith.abstract.extension import activate_ext
from nnsmith.backends.factory import BackendFactory
from nnsmith.cli.model_exec import verify_testcase
from nnsmith.error import InternalError
from nnsmith.filter import FILTERS
from nnsmith.graph_gen import model_gen
from nnsmith.logging import FUZZ_LOG
from nnsmith.macro import NNSMITH_BUG_PATTERN_TOKEN
from nnsmith.materialize import Model, TestCase
from nnsmith.narrow_spec import auto_opset
from nnsmith.util import (
    hijack_patch_requires,
    mkdir,
    op_filter,
    parse_timestr,
    set_seed,
)


class StatusCollect:
    def __init__(self, root):
        self.root = Path(root)
        mkdir(self.root)
        self.n_testcases = 0
        self.n_bugs = 0
        self.n_fail_make_test = 0

    def get_next_bug_path(self):
        return self.root / f"bug-{NNSMITH_BUG_PATTERN_TOKEN}-{self.n_bugs}"


class FuzzingLoop:
    def __init__(
        self,
        cfg: DictConfig,
    ):
        self.cfg = cfg

        # FIXME(@ganler): well-form the fix or report to TF
        # Dirty fix for TFLite on CUDA-enabled systems.
        # If CUDA is not needed, disable them all.
        cmpwith = cfg["cmp"]["with"]
        if cfg["backend"]["type"] == "tflite" and (
            cmpwith is None or cmpwith["target"] != "cuda"
        ):
            os.environ["CUDA_VISIBLE_DEVICES"] = "-1"

        if (
            cfg["fuzz"]["crash_safe"]
            and cfg["backend"]["type"] == "xla"
            and cfg["backend"]["target"] == "cuda"
        ) or (
            cmpwith is not None
            and cmpwith["type"] == "xla"
            and cmpwith["target"] == "cuda"
        ):
            raise ValueError(
                "Please set `fuzz.crash_safe=false` for XLA on CUDA. "
                "Also see https://github.com/ise-uiuc/nnsmith/blob/main/doc/known-issues.md"
            )

        self.crash_safe = bool(cfg["fuzz"]["crash_safe"])
        self.test_timeout = cfg["fuzz"]["test_timeout"]
        if self.test_timeout is not None:
            if isinstance(self.test_timeout, str):
                self.test_timeout = parse_timestr(self.test_timeout)
            assert isinstance(
                self.test_timeout, int
            ), "`fuzz.test_timeout` must be an integer (or with `s` (default), `m`/`min`, or `h`/`hr`)."

        if not self.crash_safe and self.test_timeout is not None:
            # user enabled test_timeout but not crash_safe.
            FUZZ_LOG.warning(
                "`fuzz.crash_safe` is automatically enabled given `fuzz.test_timeout` is set."
            )

        self.filters = []
        # add filters.
        filter_types = (
            [cfg["filter"]["type"]]
            if isinstance(cfg["filter"]["type"], str)
            else cfg["filter"]["type"]
        )
        if filter_types:
            patches = (
                [cfg["filter"]["patch"]]
                if isinstance(cfg["filter"]["patch"], str)
                else cfg["filter"]["patch"]
            )
            for f in patches:
                assert os.path.isfile(
                    f
                ), "filter.patch must be a list of file locations."
                assert "@filter(" in open(f).read(), f"No filter found in the {f}."
                spec = spec_from_file_location("module.name", f)
                spec.loader.exec_module(module_from_spec(spec))
                FUZZ_LOG.info(f"Imported filter patch: {f}")
            for filter in filter_types:
                filter = str(filter)
                if filter not in FILTERS:
                    raise ValueError(
                        f"Filter {filter} not found. Available filters: {FILTERS.keys()}"
                    )
                fn_or_cls = FILTERS[filter]
                if isinstance(fn_or_cls, Type):
                    self.filters.append(fn_or_cls())
                elif isinstance(fn_or_cls, FunctionType):
                    self.filters.append(fn_or_cls)
                else:
                    raise InternalError(
                        f"Invalid filter type: {fn_or_cls} (aka {filter})"
                    )
                FUZZ_LOG.info(f"Filter enabled: {filter}")

        self.status = StatusCollect(cfg["fuzz"]["root"])

        self.factory = BackendFactory.init(
            cfg["backend"]["type"],
            target=cfg["backend"]["target"],
            optmax=cfg["backend"]["optmax"],
        )

        model_cfg = self.cfg["model"]
        self.ModelType = Model.init(
            model_cfg["type"], backend_target=cfg["backend"]["target"]
        )
        self.ModelType.add_seed_setter()
        self.opset = op_filter(
            auto_opset(self.ModelType, self.factory, vulops=cfg["mgen"]["vulops"]),
            cfg["mgen"]["include"],
            cfg["mgen"]["exclude"],
        )

        hijack_patch_requires(cfg["mgen"]["patch_requires"])
        activate_ext(opset=self.opset, factory=self.factory)

        seed = cfg["fuzz"]["seed"] or random.getrandbits(32)
        set_seed(seed)

        FUZZ_LOG.info(
            f"Test success info supressed -- only showing logs for failed tests"
        )

        # Time budget checking.
        self.timeout_s = self.cfg["fuzz"]["time"]
        if isinstance(self.timeout_s, str):
            self.timeout_s = parse_timestr(self.timeout_s)
        assert isinstance(
            self.timeout_s, int
        ), "`fuzz.time` must be an integer (with `s` (default), `m`/`min`, or `h`/`hr`)."

        self.save_test = cfg["fuzz"]["save_test"]
        if isinstance(self.save_test, str):  # path of root dir.
            FUZZ_LOG.info(f"Saving all intermediate testcases to {self.save_test}")
            mkdir(self.save_test)

    def make_testcase(self, seed) -> TestCase:
        mgen_cfg = self.cfg["mgen"]
        gen = model_gen(
            opset=self.opset,
            method=mgen_cfg["method"],
            seed=seed,
            max_elem_per_tensor=mgen_cfg["max_elem_per_tensor"],
            max_nodes=mgen_cfg["max_nodes"],
            timeout_ms=mgen_cfg["timeout_ms"],
            rank_choices=mgen_cfg["rank_choices"],
            dtype_choices=mgen_cfg["dtype_choices"],
        )

        ir = gen.make_concrete()
        model = self.ModelType.from_gir(ir)
        if self.cfg["debug"]["viz"]:
            model.attach_viz(ir)

        model.refine_weights()  # either random generated or gradient-based.
        oracle = model.make_oracle()
        return TestCase(model, oracle)

    def validate_and_report(self, testcase: TestCase) -> bool:
        if not verify_testcase(
            self.cfg["cmp"],
            factory=self.factory,
            testcase=testcase,
            output_dir=self.status.get_next_bug_path(),
            filters=self.filters,
            crash_safe=self.crash_safe,
            timeout=self.test_timeout,
        ):
            self.status.n_bugs += 1
            return False
        return True

    def run(self):
        start_time = time.time()
        while time.time() - start_time < self.timeout_s:
            seed = random.getrandbits(32)
            FUZZ_LOG.debug(f"Making testcase with seed: {seed}")

            time_stat = {}

            gen_start = time.time()
            try:
                testcase = self.make_testcase(seed)
            except InternalError as e:
                raise e  # propagate internal errors
            except Exception:
                FUZZ_LOG.error(
                    f"`make_testcase` failed with seed {seed}. It can be NNSmith or Generator ({self.cfg['model']['type']}) bug."
                )
                FUZZ_LOG.error(traceback.format_exc())
                self.status.n_fail_make_test += 1
                continue
            time_stat["gen"] = time.time() - gen_start

            eval_start = time.time()
            if not self.validate_and_report(testcase):
                FUZZ_LOG.warning(f"Failed model seed: {seed}")
            time_stat["eval"] = time.time() - eval_start

            if self.save_test:
                save_start = time.time()
                testcase_dir = os.path.join(
                    self.save_test, f"{time.time() - start_time:.3f}"
                )
                mkdir(testcase_dir)
                testcase.dump(testcase_dir)
                time_stat["save"] = time.time() - save_start

            FUZZ_LOG.info(
                f"Timing: { ''.join(f'{k}: {1000 * v:.1f}ms, ' for k, v in time_stat.items()) }"
            )
            self.status.n_testcases += 1
        FUZZ_LOG.info(f"Total {self.status.n_testcases} testcases generated.")
        FUZZ_LOG.info(f"Total {self.status.n_bugs} bugs found.")
        FUZZ_LOG.info(f"Total {self.status.n_fail_make_test} failed to make testcases.")


@hydra.main(version_base=None, config_path="../config", config_name="main")
def main(cfg: DictConfig):
    FuzzingLoop(cfg).run()


if __name__ == "__main__":
    main()



import torch
from multipledispatch import dispatch

from nnsmith.abstract.op import *

# used proxy gradient functions
SLOPE = 0.1


class PGTruncFuncBase(torch.autograd.Function):
    # incomplete class
    @staticmethod
    def backward(ctx, grad_output):
        # let f' = x
        return grad_output


class PGFloorFunc(PGTruncFuncBase):
    @staticmethod
    def forward(ctx, input):
        return torch.floor(input)


class PGRoundFunc(PGTruncFuncBase):
    @staticmethod
    def forward(ctx, input):
        return torch.round(input)


class PGCeilFunc(PGTruncFuncBase):
    @staticmethod
    def forward(ctx, input):
        return torch.ceil(input)


class PGReLUFunc(torch.autograd.Function):
    @staticmethod
    def forward(ctx, input):
        ctx.save_for_backward(input)
        if input.dtype == torch.float16:
            return torch.relu(input.float()).half()
        return torch.relu(input)

    @staticmethod
    def backward(ctx, grad_output):
        (input,) = ctx.saved_tensors
        # let f' = l * x in bound
        # let f' = x out bound
        grad_input = grad_output.clone()
        return torch.where(input > 0, grad_input, grad_input * SLOPE)


class PGClipFunc(torch.autograd.Function):
    @staticmethod
    def forward(ctx, input, min, max):
        ctx.save_for_backward(input)
        ctx.clip_min = min
        ctx.clip_max = max
        return torch.clip(input, min=min, max=max)

    @staticmethod
    def backward(ctx, grad_output):
        (input,) = ctx.saved_tensors
        # let f' = l * x in bound
        # let f' = x out bound
        grad_input = grad_output.clone()
        return (
            torch.where(
                (input > ctx.clip_max).logical_or(input < ctx.clip_min),
                grad_input * SLOPE,
                grad_input,
            ),
            None,
            None,
        )


# Modules.


class PGCeil(torch.nn.Module):
    def __init__(self):
        super(PGCeil, self).__init__()

    def forward(self, x):
        return PGCeilFunc.apply(x)


class PGFloor(torch.nn.Module):
    def __init__(self):
        super(PGFloor, self).__init__()

    def forward(self, x):
        return PGFloorFunc.apply(x)


class PGRound(torch.nn.Module):
    def __init__(self):
        super(PGRound, self).__init__()

    def forward(self, x):
        return PGRoundFunc.apply(x)


class PGReLU(torch.nn.Module):
    def __init__(self):
        super(PGReLU, self).__init__()

    def forward(self, x):
        return PGReLUFunc.apply(x)


class PGClip(torch.nn.Module):
    def __init__(self, min, max):
        self.min = min
        self.max = max
        super(PGClip, self).__init__()

    def forward(self, x):
        return PGClipFunc.apply(x, self.min, self.max)


# proxy_fn:    proxy


@dispatch(ReLU)
def proxy_fn(op: ReLU):
    return PGReLU()


@dispatch(Ceil)
def proxy_fn(op: Ceil):
    return PGCeil()


# PGFloor
@dispatch(Floor)
def proxy_fn(op: Floor):
    return PGFloor()


# PGClip
@dispatch(Clip)
def proxy_fn(op: Clip):
    if op.input_like[0].dtype in DTYPE_GEN_FLOATS:
        return PGClip(-1.5, 1.5)
    else:
        return PGClip(-1, 1)


# Round
@dispatch(Round)
def proxy_fn(op: Round):
    return PGRound()



import time
from abc import ABC, abstractmethod
from typing import Dict, List, Tuple

import numpy as np
import torch

from nnsmith.abstract.op import DType
from nnsmith.materialize.torch.symbolnet import SymbolNet, random_tensor


class InputSearchBase(ABC):
    @staticmethod
    def apply_weights(net, weight_sample):
        with torch.no_grad():
            for name, param in net.named_parameters():
                param.copy_(weight_sample[name])

    def __init__(
        self, net: SymbolNet, start_inputs=None, start_weights=None, use_cuda=False
    ):
        self.net = net
        self.start_inputs = start_inputs
        self.start_weights = start_weights
        self.use_cuda = use_cuda

    @abstractmethod
    def search_one(self, start_inp, timeout_ms: int = None) -> Dict[str, torch.Tensor]:
        pass

    def search(
        self, max_time_ms: int = None, max_sample: int = 1
    ) -> Tuple[int, Dict[str, torch.Tensor]]:
        n_try = 0
        sat_inputs = None
        start_time = time.time()

        while (
            max_time_ms is None or time.time() - start_time < max_time_ms / 1000
        ) and n_try < max_sample:
            if self.start_weights is not None and n_try < len(self.start_weights):
                self.apply_weights(self.net, self.start_weights[n_try])
            else:
                weight_sample = {}
                for name, param in self.net.named_parameters():
                    weight_sample[name] = random_tensor(
                        param.shape, dtype=param.dtype, use_cuda=self.use_cuda
                    )
                self.apply_weights(self.net, weight_sample)

            if self.start_inputs is not None and n_try < len(self.start_inputs):
                cur_input = self.start_inputs[n_try]
            else:
                cur_input = self.net.get_random_inps(use_cuda=self.use_cuda)

            res = self.search_one(cur_input, max_time_ms)
            n_try += 1
            if res is not None:
                sat_inputs = res
                break

        return n_try, sat_inputs


class SamplingSearch(InputSearchBase):
    # Think about how people trivially generate inputs.
    def search_one(self, start_inp, timeout_ms: int = None) -> Dict[str, torch.Tensor]:
        with torch.no_grad():
            self.net.check_intermediate_numeric = True
            _ = self.net(**start_inp)
            if not self.net.invalid_found_last:
                return start_inp

            return None


class GradSearch(InputSearchBase):
    def search_one(self, start_inp, timeout_ms: int = None) -> Dict[str, torch.Tensor]:
        timeout_s = None if timeout_ms is None else timeout_ms / 1000
        return self.net.grad_input_gen(
            init_tensors=start_inp, use_cuda=self.use_cuda, max_time=timeout_s
        )


class PracticalHybridSearch(InputSearchBase):
    def __init__(
        self, net: SymbolNet, start_inputs=None, start_weights=None, use_cuda=False
    ):
        super().__init__(net, start_inputs, start_weights, use_cuda)

        self.differentiable = None

        if all([DType.is_float(v.dtype) for _, v in self.net.input_like.items()]):
            diff_test_inp = self.net.get_random_inps(use_cuda=self.use_cuda)
            for _, item in diff_test_inp.items():
                item.requires_grad_()
            self.net.forward(**diff_test_inp)
            self.differentiable = self.net.differentiable
        else:
            self.differentiable = False

    def search_one(self, start_inp, timeout_ms: int = None) -> Dict[str, torch.Tensor]:
        # if this model is purely differentiable -> GradSearch
        # otherwise                              -> SamplingSearch
        # FIXME: Estimate gradient (e.g., proxy gradient) for non-differentiable inputs.

        if self.differentiable:
            return GradSearch.search_one(self, start_inp, timeout_ms)
        else:
            return SamplingSearch.search_one(self, start_inp, timeout_ms)
